import pandas as pd
import numpy as np
import os.path as op
import os
import olga
#import olga.load_model as load_model
import tcrdist.olga_load_model as load_model #(USE THIS FOR PSUEDOGENE ASWELL AS FUNCTIONAL CDR3 GENERATION)
import tcrdist.olga_directed as seq_gen

import olga.generation_probability as generation_probability
import olga.generation_probability as pgen
from olga.utils import nt2aa, determine_seq_type
from tcrdist.paths import path_to_default_models
from olga.utils import gene_to_num_str
import warnings
path_to_olga_default_models = path_to_default_models
import random 

class OlgaModel:
    """
    A Class One Layer Up From Olga Classes To Help Store Configuration

    Attributes
    ----------
    chain_folder : string
        name of chain specific folder in the olga_default_models folder
    recomb_type : string
            'VDJ' or "VJ"
    """


    def __init__(self, chain_folder, recomb_type):
        """
        Sets up an OlgaModel that can be used multiple times. For instance to
        generate generation probabilities for 10K sequences.

        chain_folder : string
            'human_T_beta', 'human_T_alpha'
        recomb_type : string
            'VDJ' or "VJ"
        """

        self.chain_folder     = chain_folder
        self.recomb_type      = recomb_type
        self.generative_model = None
        self.genomic_data     = None
        self.pgen_model       = None
        self.seq_gen_model    = None 
        

        self._validate_chain_folder_arg()
        self._validate_recomb_type_arg()
        self._validate_chain_folder_with_recomb_type()



        params_file_name = op.join(path_to_olga_default_models,
                                   chain_folder,
                                   'model_params.txt')
        marginals_file_name = op.join(path_to_olga_default_models,
                                      chain_folder,
                                      'model_marginals.txt')
        V_anchor_pos_file = op.join(path_to_olga_default_models,
                                    chain_folder,
                                    'V_gene_CDR3_anchors.csv')
        J_anchor_pos_file = op.join(path_to_olga_default_models,
                                    chain_folder,
                                     'J_gene_CDR3_anchors.csv')

            #Load up model based on recomb_type
        #VDJ recomb case --- used for TCRB and IGH
        if recomb_type == 'VDJ':
            genomic_data = load_model.GenomicDataVDJ()
            genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
            generative_model = load_model.GenerativeModelVDJ()
            generative_model.load_and_process_igor_model(marginals_file_name)
            pgen_model = generation_probability.GenerationProbabilityVDJ(generative_model, genomic_data)
            self.genomic_data = genomic_data
            self.generative_model = generative_model
            self.pgen_model = pgen_model
            self.seq_gen_model = seq_gen.SequenceGenerationVDJ(self.generative_model, self.genomic_data)

        #VJ recomb case --- used for TCRA and light chain
        elif recomb_type == 'VJ':
            genomic_data = load_model.GenomicDataVJ()
            genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
            generative_model = load_model.GenerativeModelVJ()
            generative_model.load_and_process_igor_model(marginals_file_name)
            pgen_model = generation_probability.GenerationProbabilityVJ(generative_model, genomic_data)
            self.genomic_data = genomic_data
            self.generative_model = generative_model
            self.pgen_model = pgen_model
            self.seq_gen_model = seq_gen.SequenceGenerationVJ(self.generative_model, self.genomic_data)

    def __repr__(self):
        return "tcrdist.pgen.OlgaModel set to recomb_type: '{}' and chain_folder: '{}'".format(self.recomb_type, self.chain_folder)

    def gen_cdr3s(self, V:str=None,J:str=None,n:int=1)->list:
        """
        Generate n cdr3s from modifieid Olga code using a directed V and J gene usage 

        V: str
            gene name, e.g, 'TRAV27*01'
        J: str
            gene name,  e.g, 'TRAJ42*01'
        n: int 
            number of cdr3s to sample

        Returns
        -------
        cdr3s : list
            list of strings (and possibly Nones)

        Example
        -------
        np.random.seed(310)
        result = olga_model_beta.gen_cdr3s(V = 'TRBV20-1*01', J = 'TRBJ1-2*01', n = 4)
        expected = ['CSARVREAGRTYTF', 'CSAVPPGLPNYGYTF', 'CSARGPSQGYVRGLYGYTF', 'CSAQGLAGYGYTF']
        assert result == expected
        """
        cdr3s = list()
        for _ in range(n):
            r = self.gen_cdr3(V=V,J=J)
            if r is not None:
                cdr3s.append(r[1])
            else:
                cdr3s.append(None)
        return cdr3s

    def gen_cdr3(self, V=None,J=None):
        """
        Gemerate an Olga type CDR3, directed by specific gene usage 
        
        Parameters
        ----------
        V : str or Npne
        
        J : str or Npne
        
        
        Returns
        -------
        tuple

        
        Notes
        -----
         .gen_cdr3() returns the full output tuple
        
        Example
        -------
        np.random.seed(310)
        from tcrdist.pgen import OlgaModel
        olga_model_beta = OlgaModel(recomb_type="VDJ", chain_folder = "human_T_beta")
        result = olga_model_beta.gen_cdr3(V = 'TRBV20-1*01', J = 'TRBJ1-2*01')
        # NOTE: seed is set, so we expect standard result
        #NOTE: .gen_cdr3() returns the full output tuple
        expected = ('TGCAGTGCTAGAGTAAGGGAAGCGGGAAGGACCTACACCTTC',
                     'CSARVREAGRTYTF',
                     29,
                     1,
                     {'V': 29,
                      'D': 2,
                      'J': 1,
                      'delV': 5,
                      'delJ': 15,
                      'delDl': 10,
                      'delDr': 7,
                      'insVD': 7,
                      'insDJ': 6})
        assert result == expected
        """
        if V is not None:
            #print(F"V SPECIFIED {V}")
            #print(f"{V} gene_to_num_str {gene_to_num_str(V, 'V')}")
            try: 
               # print(f"pgen_model.V_mask_mapping: {self.pgen_model.V_mask_mapping[gene_to_num_str(V, 'V')]}")
                vnum = random.choice(self.pgen_model.V_mask_mapping[gene_to_num_str(V, "V")])
            except KeyError:
                warnings.warn(f"{V}:{gene_to_num_str(V, 'V')} not supported in Olga pgen_model, returning None", stacklevel=2)
                return None
        else:
            vnum = None  
            
            #print(vnum)

        if J is not None:
            #print(F"J SPECIFIED {J}")
            #print(f"{J} gene_to_num_str {gene_to_num_str(J, 'J')}")
            try:
                #print(f"pgen_model.J_mask_mapping: {self.pgen_model.J_mask_mapping[gene_to_num_str(J, 'J')]}")
                jnum = random.choice(self.pgen_model.J_mask_mapping[gene_to_num_str(J, "J")])
            except KeyError:
                warnings.warn(f"{J}:{gene_to_num_str(J, 'J')} not supported in Olga pgen_model, returning None", stacklevel=2)
                return None
        else:
            jnum = None       

        if V is not None:
            generated = self.seq_gen_model.gen_rnd_prod_CDR3(V = vnum, J= jnum)
            #if generated is not None:
            return generated
            #else:
            #    return None
        else:
            return self.seq_gen_model.gen_rnd_prod_CDR3()

        # if V is not None:
        #     return 1
        #     print("DIRECTED")
        #     print((vnum,jnum))
        #     self.seq_gen_model.gen_rnd_prod_CDR3(V = vnum, J = jnum)
        # else:
        #return self.seq_gen_model.gen_rnd_prod_CDR3()


    def compute_aa_cdr3_pgen(self, CDR3_seq, V_usage_mask_in = None, J_usage_mask_in = None):
        """Compute Pgen for the amino acid sequence CDR3_seq.

        ORIGINAL DOC STRING FROM OLGA:

        Conditioned on the V genes/alleles indicated in V_usage_mask_in and the
        J genes/alleles in J_usage_mask_in. (Examples are TCRB sequences/model)

        Parameters
        ----------
        CDR3_seq : str
            CDR3 sequence composed of 'amino acids' -- the standard amino acids,
            plus any custom symbols for an expanded codon alphabet (note the
            standard ambiguous amino acids -- B, J, X, and Z -- are included by
            default).
        V_usage_mask_in : str or list
            An object to indicate which V alleles should be considered. The default
            input is None which returns the list of all productive V alleles.
        J_usage_mask_in : str or list
            An object to indicate which J alleles should be considered. The default
            input is None which returns the list of all productive J alleles.
        print_warnings : bool
            Determines whether warnings are printed or not. Default ON.

        Returns
        -------
        pgen : float
            The generation probability (Pgen) of the sequence

        Examples
        --------
        >>> generation_probability.compute_aa_CDR3_pgen('CAWSVAPDRGGYTF')
        1.5756106696284584e-10
        >>> generation_probability.compute_aa_CDR3_pgen('CAWSVAPDRGGYTF', 'TRBV30*01', 'TRBJ1-2*01')
        1.203646865765782e-10
        >>> generation_probability.compute_aa_CDR3_pgen('CAWXXXXXXXGYTF')
        7.8102586432014974e-05
        """
        pgen_estimate = self.pgen_model.compute_aa_CDR3_pgen(CDR3_seq, V_usage_mask_in, J_usage_mask_in)
        return(pgen_estimate)

    def compute_aa_cdr3_pgens(self, CDR3_seq, V_usage_mask_in = None, J_usage_mask_in = None):
        """
        function for computing many generation probabilities. The assumption is that
        data will come from a pd.DataFrame as lists which can contain either
        strings or None

        Parameters
        ----------

        CDR3_seq : list of strings or None

        Returns
        -------
        pgens : list
            list of floats
        """
        if V_usage_mask_in is None:
            V_usage_mask_in = [None]*len(CDR3_seq)
            if len(CDR3_seq) != len(V_usage_mask_in):
                raise TypeError("len(CDR3_seq) must equal len(V_usage_mask_in)")
        else:
            if len(CDR3_seq) != len(V_usage_mask_in):
                raise TypeError("len(CDR3_seq) must equal len(V_usage_mask_in)")

        if J_usage_mask_in is None:
            J_usage_mask_in = [None]*len(CDR3_seq)
            if len(CDR3_seq) != len(J_usage_mask_in):
                raise TypeError("len(CDR3_seq) must equal len(V_usage_mask_in)")
        else:
            if len(CDR3_seq) != len(J_usage_mask_in):
                raise TypeError("len(CDR3_seq) must equal len(V_usage_mask_in)")

        # Deal with NaN
        l = len(CDR3_seq)
        input_tuples  = [(CDR3_seq[i], V_usage_mask_in[i], J_usage_mask_in[i] ) \
                         for i in range(l)]

        pgen_estimates = [self.compute_aa_cdr3_pgen(x,y,z) for x,y,z in input_tuples]
        return(pgen_estimates)



    def _validate_recomb_type_arg(self):
        ["VDJ","VJ"].index(self.recomb_type)


    def _validate_chain_folder_arg(self):
        """
        checks taht chain argument is valid based on folders in the
        path_to_olga_default_models

        Parameters
        ----------
        chain_folder : string

        Raises
        ------
        ValueError

        """
        valid_chain_folders = [x for x in os.listdir(path_to_olga_default_models)]
        valid_chain_folders_string =  " ".join(map( str, valid_chain_folders))
        try:
            valid_chain_folders.index(self.chain_folder)
        except ValueError:
            raise ValueError('chain_folder arg in OlgaModel must be: {}'.format(valid_chain_folders_string))
        return(1)

    def _validate_chain_folder_with_recomb_type(self):
        """
        check for correct recomb_type cell type match

        TODO: Add DELTA GAMMA and MOUSE RULES
        """
        if self.recomb_type == "VDJ" and self.chain_folder == "human_T_alpha":
            raise ValueError("human_T_alpha default is recomb_type = VJ")
        if self.recomb_type == "VJ" and self.chain_folder == "human_T_beta":
            raise ValueError("human_T_beta default is recomb_type = VDJ")
        return(1)



# 




#
#
#
#
# def olga_pgen(chain_folder = 'human_T_beta',
#               cdr3 = 'CAWSVAPDRGGYTF',
#               v_b = 'TRBV30*01',
#               v_j ='TRBJ1-2*01'):
#     chain_folder = 'human_T_beta'
#     params_file_name = op.join(path_to_olga_default_models,
#                                chain_folder,
#                                'model_params.txt')
#     marginals_file_name = op.join(path_to_olga_default_models,
#                                   chain_folder,
#                                   'model_marginals.txt')
#     V_anchor_pos_file = op.join(path_to_olga_default_models,
#                                 chain_folder,
#                                 'V_gene_CDR3_anchors.csv')
#     J_anchor_pos_file = op.join(path_to_olga_default_models,
#                                 chain_folder,
#                                  'J_gene_CDR3_anchors.csv')
#
#     #Load data
#     genomic_data = load_model.GenomicDataVDJ()
#     genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
#     #Load model
#     generative_model = load_model.GenerativeModelVDJ()
#     generative_model.load_and_process_igor_model(marginals_file_name)
#
#     #Process model/data for pgen computation by instantiating GenerationProbabilityVDJ
#     pgen_model = pgen.GenerationProbabilityVDJ(generative_model, genomic_data)
#
#     #Compute some sequence pgens
#     x = pgen_model.compute_aa_CDR3_pgen('CAWSVAPDRGGYTF', 'TRBV30*01', 'TRBJ1-2*01')
#     assert(np.isclose(x,1.203646865765782e-10))
#     return(x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Define classes for V(D)J generative models and genomic data.

    Copyright (C) 2018 Zachary Sethna

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
    
    
The classes GenomicDataVDJ and GenomicDataVJ are used to format the germline
genomic sequences/data.

The classes GenerativeModelVDJ and GenerativeModelVJ are used to format V(D)J
recombination model parameters.

Normally the genomic data and model parameters are being read from IGoR 
inference files, and V and J anchor files that have been prepared. For this
purpose methods are included into these classes to read in IGoR files and set
the relevant attributes based on the parameters read in. 

IF YOU DO NOT WANT TO USE IGOR INFERENCE FILES (OR MATCH THEIR SYNTAX), YOU 
WILL NEED WRITE A METHOD THAT WILL PARSE SOME OTHER FILE SYNTAX.

For GenomicDataVDJ and GenomicDataVJ the method:

genomic_data.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)

will read in the IGoR parameter file params_file_name for the gene/allele
names and germline sequences. The V_anchor_pos_file and J_anchor_pos_file
indicate where in the germline sequences the conserved residue that is used to
define the CDR3 sequence (in general a cysteine 'C' for the V and a 
phenylalanine 'F' or a tryptophan 'W' for the J) as well as whether the V or J
gene/allele is a functional gene or a pseudo gene. It must be noted that the 
gene/allele names from params_file_name MUST EXACTLY MATCH the names that can 
be found in V_anchor_pos_file and J_anchor_pos_file (the exact strings must
match) for the files to load properly. If you are using an IGoR inference file
that is not one of the default models shipped, be sure that the anchor files
match your IGoR parameter file. 


For GenerativeModelVDJ and GenerativeModelVJ:

generative_model.load_and_process_igor_model(marginals_file_name)

reads and processes an IGoR marginals file to construct the probability
distributions of a generative V(D)J model. The assumed model factorization of
the models are the same as in the paper:

VDJ model:
    
P(event) = P(V)P(delV|V)P(D,J)P(delDl,delDr|D)P(delJ|J)P(m_1...m_L)P(n_K...n_1)
where the insertion junction probabilities are a dinucleotide Markov model:
P(m_1...m_L) = PinsVD(L) p_0(m_1) \prod_{i=2}^L Rvd(m_i|m_{i-1})
P(n_K...n_1) = PinsDJ(K) q_0(n_1) \prod_{i=2}^K Rdj(n_i|n_{i-1})

VJ model:
P(event) = P(V,J)P(delV|V)P(delJ|J)P(m_1...m_L)
where the insertion junction is again a dinucleotide Markov model:
P(m_1...m_L) = PinsVJ(L) p_0(m_1) \prod_{i=2}^L Rvj(m_i|m_{i-1})    

The attributes of GenerativeModel (and subsequent processed attributes in 
other classes) attempt to preserve the names from the model factorization as
closely as possible (and thus do not follow the proper Python etiquette of all
lower case with underscores for variable/attribute names).


Some of the functions that directly read in IGoR parameter/marginal files
are adapted from Quentin Marcou (author of IGoR).

@author: zacharysethna

"""
from __future__ import print_function, division
import numpy as np
from olga.utils import cutR_seq, cutL_seq, calc_S_single_gene, calc_S_joint_genes, calc_Sins

#%% GenomicData class definitions
class GenomicData(object):
    """Class used to load genomic data that both VDJ and VJ models have.
    
    The class is the parent of classes GenomicDataVDJ and 
    GenomicDataVJ.

    Attributes
    ----------
    genV : list of lists
        List of genomic V information.
    max_delV_palindrome : int
        Maximum number of reverse complementary insertions for a V germline
        sequence.           
    cutV_genomic_CDR3_segs : list of strings
        List of the V germline nucleotide sequences, trimmed to begin at the
        CDR3 region (includes the conserved C residue) with the maximum number
        of reverse complementary palindromic insertions appended.
        
    genJ : list  of lists
        List of genomic J information.       
    max_delJ_palindrome : int
        Maximum number of reverse complementary insertions for a J germline
        sequence
    cutJ_genomic_CDR3_segs : list of strings
        List of the J germline nucleotide sequences, trimmed to end at the
        CDR3 region (includes the conserved F or W residue) with the maximum 
        number of reverse complementary palindromic insertions appended.
        
    """
    def __init__(self):
        """Initialize GenomicData."""
        
        self.genV = None
        self.genJ = None
        
        self.max_delV_palindrome = None
        self.max_delJ_palindrome = None
        
        self.cutV_genomic_CDR3_segs = None
        self.cutJ_genomic_CDR3_segs = None
        
    def anchor_and_curate_genV_and_genJ(self, V_anchor_pos_file, J_anchor_pos_file):
        """Trim V and J germline sequences to the CDR3 region.
        
        Unproductive sequences have an empty string '' for the CDR3 region
        sequence.
        
        Edits the attributes genV and genJ
        
        Parameters
        ----------
        V_anchor_pos_file_name : str
            File name for the conserved residue (C) locations and functionality 
            of each V genomic sequence.
        J_anchor_pos_file_name : str
            File name for the conserved residue (F/W) locations and 
            functionality of each J genomic sequence.
        
        """
        
        V_anchor_pos = load_genomic_CDR3_anchor_pos_and_functionality(V_anchor_pos_file)
        J_anchor_pos = load_genomic_CDR3_anchor_pos_and_functionality(J_anchor_pos_file)
        
        for V in self.genV:
            try:
                if V_anchor_pos[V[0]][0] > 0 and V_anchor_pos[V[0]][1] != 'X': #Check for functionality, here we allow P
                    V[1] = V[2][V_anchor_pos[V[0]][0]:]
                else:
                    V[1] = ''
            except KeyError:
                V[1] = ''
    
        for J in self.genJ:
            try:
                if J_anchor_pos[J[0]][0] > 0 and J_anchor_pos[J[0]][1] != 'X': #Check for functionality, here we allow P
                    J[1] = J[2][:J_anchor_pos[J[0]][0]+3]
                else:
                    J[1] = ''
            except KeyError:
                J[1] = ''
        
    def generate_cutV_genomic_CDR3_segs(self):
        """Add palindromic inserted nucleotides to germline V sequences.
        
        The maximum number of palindromic insertions are appended to the
        germline V segments so that delV can index directly for number of
        nucleotides to delete from a segment.
        
        Sets the attribute cutV_genomic_CDR3_segs.
        
        """
    
        max_palindrome = self.max_delV_palindrome

        self.cutV_genomic_CDR3_segs = []
        for CDR3_V_seg in [x[1] for x in self.genV]:
            if len(CDR3_V_seg) < max_palindrome:
                self.cutV_genomic_CDR3_segs += [cutR_seq(CDR3_V_seg, 0, len(CDR3_V_seg))]
            else:
                self.cutV_genomic_CDR3_segs += [cutR_seq(CDR3_V_seg, 0, max_palindrome)]
                
    def generate_cutJ_genomic_CDR3_segs(self):
        """Add palindromic inserted nucleotides to germline J sequences.
        
        The maximum number of palindromic insertions are appended to the
        germline J segments so that delJ can index directly for number of
        nucleotides to delete from a segment.
        
        Sets the attribute cutJ_genomic_CDR3_segs.
        
        """
        
        max_palindrome = self.max_delJ_palindrome
        self.cutJ_genomic_CDR3_segs = []
        for CDR3_J_seg in [x[1] for x in self.genJ]:
            if len(CDR3_J_seg) < max_palindrome:
                self.cutJ_genomic_CDR3_segs += [cutL_seq(CDR3_J_seg, 0, len(CDR3_J_seg))]
            else:
                self.cutJ_genomic_CDR3_segs += [cutL_seq(CDR3_J_seg, 0, max_palindrome)]

        
class GenomicDataVDJ(GenomicData):
    """Class used to load genomic data of a VDJ model.

    Attributes
    ----------
    genV : list of lists
        List of genomic V information.
    max_delV_palindrome : int
        Maximum number of reverse complementary insertions for a V germline
        sequence.           
    cutV_genomic_CDR3_segs : list of strings
        List of the V germline nucleotide sequences, trimmed to begin at the
        CDR3 region (includes the conserved C residue) with the maximum number
        of reverse complementary palindromic insertions appended.
        
    genD : list of lists
        List of genomic D information.
    max_delDl_palindrome : int
        Maximum number of reverse complementary insertions for a D germline
        sequence from the left (5').
    max_delDr_palindrome : int
        Maximum number of reverse complementary insertions for a D germline
        sequence from the right (3').
    cutD_genomic_CDR3_segs : list of strings
        List of the D germline nucleotide sequences, with the maximum number
        of reverse complementary palindromic insertions appended to both ends.
        
    genJ : list  of lists
        List of genomic J information.       
    max_delJ_palindrome : int
        Maximum number of reverse complementary insertions for a J germline
        sequence
    cutJ_genomic_CDR3_segs : list of strings
        List of the J germline nucleotide sequences, trimmed to end at the
        CDR3 region (includes the conserved F or W residue) with the maximum 
        number of reverse complementary palindromic insertions appended.
        
    """
    
    def __init__(self, params_file_name = None, V_anchor_pos_file = None, J_anchor_pos_file = None):
        """Initialize GenomicDataVDJ.
        
        Parameters
        ----------
        params_file_name : str or None
            File name for a IGOR parameter file.
        V_anchor_pos_file_name : str or None
            File name for the conserved residue (C) locations and functionality 
            of each V genomic sequence.
        J_anchor_pos_file_name : str or None
            File name for the conserved residue (F/W) locations and 
            functionality of each J genomic sequence.
        
        """
        GenomicData.__init__(self)
        
        self.genD = None
        self.delDl_palindrome = None
        self.delDr_palindrome = None 
        self.cutD_genomic_CDR3_segs = None
        
        if all([params_file_name is not None, V_anchor_pos_file is not None, J_anchor_pos_file is not None]):
            self.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
        
    
    def load_igor_genomic_data(self, params_file_name, V_anchor_pos_file, J_anchor_pos_file):
        """Set attributes by loading in genomic data from IGoR parameter file.
        
        Sets attributes genV,  max_delV_palindrome, cutV_genomic_CDR3_segs, 
        genD, max_delDl_palindrome, max_delDr_palindrome, 
        cutD_genomic_CDR3_segs, genJ, max_delJ_palindrome, and 
        cutJ_genomic_CDR3_segs.
        
        Parameters
        ----------
        params_file_name : str
            File name for a IGOR parameter file.
        V_anchor_pos_file_name : str
            File name for the conserved residue (C) locations and functionality 
            of each V genomic sequence.
        J_anchor_pos_file_name : str
            File name for the conserved residue (F/W) locations and 
            functionality of each J genomic sequence.
        
        """
        
        self.genV = read_igor_V_gene_parameters(params_file_name)
        self.genD = read_igor_D_gene_parameters(params_file_name)
        self.genJ = read_igor_J_gene_parameters(params_file_name)
        
        self.anchor_and_curate_genV_and_genJ(V_anchor_pos_file, J_anchor_pos_file)
 
        self.read_VDJ_palindrome_parameters(params_file_name) #Need palindrome info before generating cut_genomic_CDR3_segs

        self.generate_cutV_genomic_CDR3_segs()
        self.generate_cutD_genomic_CDR3_segs()
        self.generate_cutJ_genomic_CDR3_segs()
                
    def generate_cutD_genomic_CDR3_segs(self):
        """Add palindromic inserted nucleotides to germline V sequences.
        
        The maximum number of palindromic insertions are appended to the
        germline D segments so that delDl and delDr can index directly for number 
        of nucleotides to delete from a segment.
        
        Sets the attribute cutV_genomic_CDR3_segs.
        
        """
        max_palindrome_L = self.max_delDl_palindrome
        max_palindrome_R = self.max_delDr_palindrome

        self.cutD_genomic_CDR3_segs = []
        for CDR3_D_seg in [x[1] for x in self.genD]:
            if len(CDR3_D_seg) < min(max_palindrome_L, max_palindrome_R):
                self.cutD_genomic_CDR3_segs += [cutR_seq(cutL_seq(CDR3_D_seg, 0, len(CDR3_D_seg)), 0, len(CDR3_D_seg))]
            else:
                self.cutD_genomic_CDR3_segs += [cutR_seq(cutL_seq(CDR3_D_seg, 0, max_palindrome_L), 0, max_palindrome_R)]
    
    def read_VDJ_palindrome_parameters(self, params_file_name):
        """Read V, D, and J palindrome parameters from file.
        
        Sets the attributes max_delV_palindrome, max_delDl_palindrome,
        max_delDr_palindrome, and max_delJ_palindrome.
    
        Parameters
        ----------
        params_file_name : str
            File name for an IGoR parameter file of a VDJ generative model.
        
        """
        
        params_file = open(params_file_name, 'r')
        
        
        in_delV = False
        in_delDl = False
        in_delDr = False
        in_delJ = False
        
        
        for line in params_file:
            if line.startswith('#Deletion;V_gene;'):
                in_delV = True
                in_delDl = False
                in_delDr = False
                in_delJ = False
            elif line.startswith('#Deletion;D_gene;Three_prime;'):
                in_delV = False
                in_delDl = False
                in_delDr = True
                in_delJ = False
            elif line.startswith('#Deletion;D_gene;Five_prime;'):
                in_delV = False
                in_delDl = True
                in_delDr = False
                in_delJ = False
            elif line.startswith('#Deletion;J_gene;'):
                in_delV = False
                in_delDl = False
                in_delDr = False
                in_delJ = True
            elif any([in_delV, in_delDl, in_delDr, in_delJ]) and line.startswith('%'):
                if int(line.split(';')[-1]) == 0:
                    if in_delV:
                        self.max_delV_palindrome = np.abs(int(line.lstrip('%').split(';')[0]))
                    elif in_delDl:
                        self.max_delDl_palindrome = np.abs(int(line.lstrip('%').split(';')[0]))
                    elif in_delDr:
                        self.max_delDr_palindrome = np.abs(int(line.lstrip('%').split(';')[0]))
                    elif in_delJ:
                        self.max_delJ_palindrome = np.abs(int(line.lstrip('%').split(';')[0]))
            else:
                in_delV = False
                in_delDl = False
                in_delDr = False
                in_delJ = False
                
                
class GenomicDataVJ(GenomicData):
    """Class used to load genomic data of a VJ model.

    Attributes
    ----------
    genV : list of lists
        List of genomic V information.
    max_delV_palindrome : int
        Maximum number of reverse complementary insertions for a V germline
        sequence.           
    cutV_genomic_CDR3_segs : list of strings
        List of the V germline nucleotide sequences, trimmed to begin at the
        CDR3 region (includes the conserved C residue) with the maximum number
        of reverse complementary palindromic insertions appended.
        
    genJ : list  of lists
        List of genomic J information.       
    max_delJ_palindrome : int
        Maximum number of reverse complementary insertions for a J germline
        sequence
    cutJ_genomic_CDR3_segs : list of strings
        List of the J germline nucleotide sequences, trimmed to end at the
        CDR3 region (includes the conserved F or W residue) with the maximum 
        number of reverse complementary palindromic insertions appended.
        
    """
    def __init__(self, params_file_name = None, V_anchor_pos_file = None, J_anchor_pos_file = None):
        """Initialize GenomicDataVJ.
        
        Parameters
        ----------
        params_file_name : str or None
            File name for a IGOR parameter file.
        V_anchor_pos_file_name : str or None
            File name for the conserved residue (C) locations and functionality 
            of each V genomic sequence.
        J_anchor_pos_file_name : str or None
            File name for the conserved residue (F/W) locations and 
            functionality of each J genomic sequence.
        
        """
        GenomicData.__init__(self)
        if all([params_file_name is not None, V_anchor_pos_file is not None, J_anchor_pos_file is not None]):
            self.load_igor_genomic_data(params_file_name, V_anchor_pos_file, J_anchor_pos_file)
        
    def load_igor_genomic_data(self, params_file_name, V_anchor_pos_file, J_anchor_pos_file):
        """Set attributes by loading in genomic data from IGoR parameter file.
        
        Sets attributes genV, genJ, max_delV_palindrome, max_delJ_palindrome,
        cutV_genomic_CDR3_segs, and cutJ_genomic_CDR3_segs.
        
        Parameters
        ----------
        params_file_name : str
            File name for a IGOR parameter file.
        V_anchor_pos_file_name : str
            File name for the conserved residue (C) locations and functionality 
            of each V genomic sequence.
        J_anchor_pos_file_name : str
            File name for the conserved residue (F/W) locations and 
            functionality of each J genomic sequence.
        
        """
        
        self.genV = read_igor_V_gene_parameters(params_file_name)
        self.genJ = read_igor_J_gene_parameters(params_file_name)
        
        self.anchor_and_curate_genV_and_genJ(V_anchor_pos_file, J_anchor_pos_file)
        
        self.read_igor_VJ_palindrome_parameters(params_file_name)
        
        self.generate_cutV_genomic_CDR3_segs()
        self.generate_cutJ_genomic_CDR3_segs()

            
    def read_igor_VJ_palindrome_parameters(self, params_file_name):
        """Read V and J palindrome parameters from file.
        
        Sets the attributes max_delV_palindrome and max_delJ_palindrome.
    
        Parameters
        ----------
        params_file_name : str
            File name for an IGoR parameter file of a VJ generative model.
        
        """
        params_file = open(params_file_name, 'r')
        
        
        in_delV = False
        in_delJ = False
        
        
        for line in params_file:
            if line.startswith('#Deletion;V_gene;'):
                in_delV = True
                in_delJ = False
            elif line.startswith('#Deletion;J_gene;'):
                in_delV = False
                in_delJ = True
            elif any([in_delV, in_delJ]) and line.startswith('%'):
                if int(line.split(';')[-1]) == 0:
                    if in_delV:
                        self.max_delV_palindrome = np.abs(int(line.lstrip('%').split(';')[0]))
                    elif in_delJ:
                        self.max_delJ_palindrome = np.abs(int(line.lstrip('%').split(';')[0]))
            else:
                in_delV = False
                in_delJ = False
                
#%% Functions for germline sequence load 
#(assumes IGoR files, and the anchor .csv files)
def load_genomic_CDR3_anchor_pos_and_functionality(anchor_pos_file_name):
    """Read anchor position and functionality from file.

    Parameters
    ----------
    anchor_pos_file_name : str
        File name for the functionality and position of a conserved residue 
        that defines the CDR3 region for each V or J germline sequence.
        
    Returns
    -------
    anchor_pos_and_functionality : dict
        Residue anchor position and functionality for each gene/allele.
    
    """
    
    anchor_pos_and_functionality = {}
    anchor_pos_file = open(anchor_pos_file_name, 'r')
    
    first_line = True
    for line in anchor_pos_file:
        if first_line:
            first_line = False
            continue
        
        split_line = line.split(',')
        split_line = [x.strip() for x in split_line]
        anchor_pos_and_functionality[split_line[0]] = [int(split_line[1]), split_line[2].strip().strip('()')]

    return anchor_pos_and_functionality
                
def read_igor_V_gene_parameters(params_file_name):
    """Load raw genV from file.
    
    genV is a list of genomic V information. Each element is a list of three 
    elements. The first is the name of the V allele, the second is the genomic 
    sequence trimmed to the CDR3 region for productive sequences, and the last 
    is the full germline sequence. For this 'raw genV' the middle element is an
    empty string to be filled in later.

    Parameters
    ----------
    params_file_name : str
        File name for a IGOR parameter file.

    Returns
    -------
    genV : list
        List of genomic V information.
    
    """
    params_file = open(params_file_name, 'r')
    
    V_gene_info = {}

    in_V_gene_sec = False
    for line in params_file:
        if line.startswith('#GeneChoice;V_gene;'):
            in_V_gene_sec = True
        elif in_V_gene_sec:
            if line[0] == '%':
                split_line = line[1:].split(';')
                V_gene_info[split_line[0]] = [split_line[1] , int(split_line[2])]
            else:
                break
    params_file.close()
    
    genV = [[]]*len(V_gene_info.keys())
    
    for V_gene in V_gene_info.keys():
        genV[V_gene_info[V_gene][1]] = [V_gene, '', V_gene_info[V_gene][0]]

    return genV

def read_igor_D_gene_parameters(params_file_name):
    """Load genD from file.
    
    genD is a list of genomic D information. Each element is a list of the name
    of the D allele and the germline sequence.

    Parameters
    ----------
    params_file_name : str
        File name for a IGOR parameter file.

    Returns
    -------
    genD : list
        List of genomic D information.
    
    """
    params_file = open(params_file_name, 'r')
    
    D_gene_info = {}

    in_D_gene_sec = False
    for line in params_file:
        if line.startswith('#GeneChoice;D_gene;'):
            in_D_gene_sec = True
        elif in_D_gene_sec:
            if line[0] == '%':
                split_line = line[1:].split(';')
                D_gene_info[split_line[0]] = [split_line[1] , int(split_line[2])]
            else:
                break
    params_file.close()
    
    genD = [[]]*len(D_gene_info.keys())
    
    for D_gene in D_gene_info.keys():
        genD[D_gene_info[D_gene][1]] = [D_gene, D_gene_info[D_gene][0]]

    return genD

def read_igor_J_gene_parameters(params_file_name):
    """Load raw genJ from file.
    
    genJ is a list of genomic J information. Each element is a list of three 
    elements. The first is the name of the J allele, the second is the genomic 
    sequence trimmed to the CDR3 region for productive sequences, and the last 
    is the full germline sequence. For this 'raw genJ' the middle element is an
    empty string to be filled in later.

    Parameters
    ----------
    params_file_name : str
        File name for a IGOR parameter file.

    Returns
    -------
    genJ : list
        List of genomic J information.
    
    """
    params_file = open(params_file_name, 'r')
    
    J_gene_info = {}

    in_J_gene_sec = False
    for line in params_file:
        if line.startswith('#GeneChoice;J_gene;'):
            in_J_gene_sec = True
        elif in_J_gene_sec:
            if line[0] == '%':
                split_line = line[1:].split(';')
                J_gene_info[split_line[0]] = [split_line[1] , int(split_line[2])]
            else:
                break
    params_file.close()
    
    genJ = [[]]*len(J_gene_info.keys())
    
    for J_gene in J_gene_info.keys():
        genJ[J_gene_info[J_gene][1]] = [J_gene, '', J_gene_info[J_gene][0]]

    return genJ

#%% GenerativeModel Class definitions

class GenerativeModelVDJ(object):
    """Class of a VDJ generative model.
    
    Attributes
    ----------
    PV : ndarray
        Probability distribution of V usage
    PdelV_given_V : ndarray
        Conditional distribution of the V deletions given the V allele,
        i.e. P(delV | V)
    
    PDJ : ndarray
        Joint probability distribution of the D and J usages.
    PdelJ_given_J : ndarray
        Conditional distribution of the J deletions given the J allele,
        i.e. P(delJ | J)
    PdelDldelDr_given_D : ndarray
        Joint probability distribution of the D deletions given the D allele,
        i.e. P(delDl, delDr |D)
    
    PinsVD : ndarray
        Probability distribution of the VD (N1) insertion sequence length        
    PinsDJ : ndarray
        Probability distribution of the DJ (N2) insertion sequence length        
    Rvd : ndarray
        Markov transition matrix for the VD insertion junction.
    Rdj : ndarray
        Markov transition matrix for the DJ insertion junction.        
    first_nt_bias_insVD : ndarray
        (4,) array of the probability distribution of the indentity of the 
        first nucleotide insertion for the VD junction.
    first_nt_bias_insDJ : ndarray
        (4,) array of the probability distribution of the indentity of the 
        first nucleotide insertion for the DJ junction.
    SV : float
        Entropy of PV (in bits).
    SdelV : float
        Conditional entropy of PdelV_givenV (in bits).
    SDJ : float
        Entropy of PDJ (in bits).
    SdelD : float
        Conditional entropy of PdelDldelDr_given_D (in bits).
    SdelJ : float
        Conditional entropy of PdelJ_given_J (in bits)
    SinsVD : float
        Entropy of of PinsVD (in bits)
    S_n1_markov : float
        Entropy of the choice of nucleotides at the N1 (VD) junction (in bits).    
    SinsDJ : float
        Entropy of PinsDJ (in bits)
    S_n2_markov : float
        Entropy of the choice of nucleotides at the N2 (DJ) junction (in bits). 
    Sscenario : float
        Total entropy of recombination event scenarios (in bits).
    
    """
    def __init__(self, marginals_file_name = None):
        """Initialize GenerativeModelVDJ.
        
        Parameters
        ----------
        marginals_file_name : str or None
            File name for a IGoR model marginals file.
            
        """
        
        self.PV = None
        self.PinsVD = None
        self.PinsDJ = None
        self.PdelV_given_V = None
        self.PdelJ_given_J = None
        self.PDJ = None
        self.PdelDldelDr_given_D = None
        self.Rvd = None
        self.Rdj = None
        
        #Normal IGoR inference does not infer these parameters, but allow for
        #the formal model
        self.first_nt_bias_insVD = None
        self.first_nt_bias_insDJ = None
        
        self.SV = None
        self.SdelV = None
        self.SDJ = None
        self.SdelD = None
        self.SdelJ = None
        self.SinsVD = None
        self.S_n1_markov = None
        self.SinsDJ = None
        self.S_n2_markov = None
        
        self.Sscenario = None
        
        if marginals_file_name is not None: 
            self.load_and_process_igor_model(marginals_file_name)
        
    def load_and_process_igor_model(self, marginals_file_name):
        """Set attributes by reading a generative model from IGoR marginal file.
        
        Sets attributes PV, PdelV_given_V, PDJ, PdelJ_given_J, 
        PdelDldelDr_given_D, PinsVD, PinsDJ, Rvd, and Rdj.
        Computes entropies SV, SdelV, SDJ, SdelD, SdelJ, SinsVD, S_n1_markov, 
        SinsDJ, S_n2_markov, and Sscenario.
        
        Parameters
        ----------
        marginals_file_name : str
            File name for a IGoR model marginals file.
        
        """
        
        
        raw_model = read_igor_marginals_txt(marginals_file_name)
        
        self.PV = raw_model[0]['v_choice']
        self.PinsVD = raw_model[0]['vd_ins']
        self.PinsDJ = raw_model[0]['dj_ins']
        self.PdelV_given_V = raw_model[0]['v_3_del'].T
        self.PdelJ_given_J = raw_model[0]['j_5_del'].T
        
        #While this class assumes P(V, D, J) factorizes into P(V)*P(D, J), the B cell model
        #infers allowing for the full correlation. Most of the correlation information is due to
        #chromosomal correlation of alleles (i.e. what chromosome each allele is found on).
        #While this information can be interesting for inference purposes, it is discarded here
        #as generally these models may be use for CDR3s from individuals the models weren't inferred
        #from (and thus the chromosomal correlations are incorrect). This also equates the T and B cell
        #models. To reintroduce the chromosomal correlations use V and J usage masks after inferring the 
        #allele identities on each chromosome.
        
        if raw_model[1]['d_gene'] == ['j_choice', 'd_gene']:
        #Factorized P(V, D, J) = P(V)*P(D, J) --- correct for T cell models
            self.PDJ = np.multiply(raw_model[0]['d_gene'].T, raw_model[0]['j_choice'])
        elif raw_model[1]['d_gene'] == ['v_choice', 'j_choice', 'd_gene']:
        #Full P(V, D, J) for B cells --- need to compute the marginal P(D, J)
            PVJ = np.multiply(raw_model[0]['j_choice'].T, raw_model[0]['v_choice']).T
            PVDJ = np.zeros([raw_model[0]['d_gene'].shape[0], raw_model[0]['d_gene'].shape[2], raw_model[0]['d_gene'].shape[1]])
            for v_in in range(raw_model[0]['d_gene'].shape[0]):
                for j_in in range(raw_model[0]['d_gene'].shape[1]):
                    PVDJ[v_in, :, j_in] = PVJ[v_in, j_in]*raw_model[0]['d_gene'][v_in, j_in, :]
            self.PDJ = np.sum(PVDJ, 0)
        else:
            print('Unrecognized model structure -- need to construct P(D, J)')
            return 0
        
        self.PdelDldelDr_given_D = np.transpose(np.multiply(np.transpose(raw_model[0]['d_3_del'], (2, 0, 1)), raw_model[0]['d_5_del']), (2, 0 , 1))
        Rvd_raw = raw_model[0]['vd_dinucl'].reshape((4, 4)).T
        self.Rvd = np.multiply(Rvd_raw, 1/np.sum(Rvd_raw, axis = 0))
        Rdj_raw = raw_model[0]['dj_dinucl'].reshape((4, 4)).T
        self.Rdj = np.multiply(Rdj_raw, 1/np.sum(Rdj_raw, axis = 0))
        
        
        #Compute Entropies
        SV, SdelV = calc_S_single_gene(self.PV, self.PdelV_given_V)
        SDJ, SdelD, SdelJ = calc_S_joint_genes(self.PDJ, self.PdelDldelDr_given_D, self.PdelJ_given_J)
        
        #Current IGoR models don't have first nt biases -- use default of steady state
        SinsVD, S_n1_markov = calc_Sins(self.PinsVD, self.Rvd)
        SinsDJ, S_n2_markov = calc_Sins(self.PinsDJ, self.Rdj)
        
        self.SV = SV
        self.SdelV = SdelV
        self.SDJ = SDJ
        self.SdelD = SdelD
        self.SdelJ = SdelJ
        self.SinsVD = SinsVD
        self.S_n1_markov = S_n1_markov
        self.SinsDJ = SinsDJ
        self.S_n2_markov = S_n2_markov
        
        self.Sscenario = SV + SdelV + SDJ + SdelD + SdelJ + SinsVD + S_n1_markov + SinsDJ + S_n2_markov
        
class GenerativeModelVJ(object):
    """Class of a VJ generative model.
    
    Attributes
    ----------
    PVJ : ndarray
        Joint probability distribution of the V and J usages.
    PdelV_given_V : ndarray
        Conditional distribution of the V deletions given the V allele,
        i.e. P(delV | V)
    PdelJ_given_J : ndarray
        Conditional distribution of the J deletions given the J allele,
        i.e. P(delJ | J)
        
    PinsVJ : ndarray
        Probability distribution of the VJ (N) insertion sequence length              
    Rvj : ndarray
        Markov transition matrix for the VJ insertion junction.       
    first_nt_bias_insVJ : ndarray
        (4,) array of the probability distribution of the indentity of the 
        first nucleotide insertion for the VD junction.
    SVJ : float
        Entropy of PVJ (in bits).
    SdelV : float
        Conditional entropy of PdelV_givenV (in bits).
    SdelJ : float
        Conditional entropy of PdelJ_given_J (in bits).
    SinsVJ : float
        Entropy of PinsVJ (in bits).
    S_n_markov : float
        Entropy of the choice of nucleotides at the N (VJ) junction (in bits).    
    Sscenario : float
        Total entropy of recombination event scenarios (in bits).
    
    """
    def __init__(self, marginals_file_name = None):
        """Initialize GenerativeModelVJ.
        
        Parameters
        ----------
        marginals_file_name : str or None
            File name for a IGoR model marginals file.
            
        """
        self.PVJ = None
        self.PinsVJ = None
        self.PdelV_given_V = None
        self.PdelJ_given_J = None
        self.Rvj = None
        
        #Normal IGoR inference does not infer these parameters, but allow for
        #the formal model
        self.first_nt_bias_insVJ = None
        
        self.SVJ = None
        self.SdelV = None
        self.SdelJ = None
        self.SinsVJ = None
        self.S_n_markov = None
        
        self.Sscenario = None

        if marginals_file_name is not None: 
            self.load_and_process_igor_model(marginals_file_name)

        
        
    def load_and_process_igor_model(self, marginals_file_name):
        """Set attributes by reading a generative model from IGoR marginal file.
        
        Sets attributes PVJ, PdelV_given_V, PdelJ_given_J, PinsVJ, and Rvj.
        
        Parameters
        ----------
        marginals_file_name : str
            File name for a IGoR model marginals file.
        
        """
        
        raw_model = read_igor_marginals_txt(marginals_file_name)
        
        self.PinsVJ = raw_model[0]['vj_ins']
        self.PdelV_given_V = raw_model[0]['v_3_del'].T
        self.PdelJ_given_J = raw_model[0]['j_5_del'].T
        self.PVJ = np.multiply( raw_model[0]['j_choice'].T, raw_model[0]['v_choice']).T
        Rvj_raw = raw_model[0]['vj_dinucl'].reshape((4, 4)).T
        self.Rvj = np.multiply(Rvj_raw, 1/np.sum(Rvj_raw, axis = 0))
        
        
        #Compute Entropies
        SVJ, SdelV, SdelJ = calc_S_joint_genes(self.PVJ, self.PdelV_given_V, self.PdelJ_given_J)
        SinsVJ, S_n_markov = calc_Sins(self.PinsVJ, self.Rvj)
        
        self.SVJ = SVJ
        self.SdelV = SdelV
        self.SdelJ = SdelJ
        self.SinsVJ = SinsVJ
        self.S_n_markov = S_n_markov
        
        self.Sscenario = SVJ + SdelV + SdelJ + SinsVJ + S_n_markov

#%% Function for reading in IGoR marginal files        
def read_igor_marginals_txt(marginals_file_name , dim_names=False):
    """Load raw IGoR model marginals.
    
    Parameters
    ----------
    marginals_file_name : str
        File name for a IGOR model marginals file.

    Returns
    -------
    model_dict : dict
        Dictionary with model marginals.
    dimension_names_dict : dict
        Dictionary that defines IGoR model dependecies.
    
    """
    with open(marginals_file_name,'r') as file:
        #Model parameters are stored inside a dictionary of ndarrays
        model_dict = {}
        dimension_names_dict = {}
        element_name=""
        first = True
        first_dim_line = False
        element_marginal_array = []
        indices_array = []

        for line in file:
            strip_line = line.rstrip('\n') #Remove end of line character
            if strip_line[0]=='@':
                first_dim_line = True
                if not(first):
                    #Add the previous to the dictionnary
                    model_dict[element_name] = element_marginal_array
                else:
                    first = False
				
                element_name = strip_line[1:]

            if strip_line[0]=='$':
                #define array dimensions
                coma_index = strip_line.find(',')
                dimensions = []

                #Get rid of $Dim[
                previous_coma_index = 4
                while coma_index != -1:
                    dimensions.append(int(strip_line[previous_coma_index+1:coma_index]))
                    previous_coma_index = coma_index
                    coma_index = strip_line.find(',',coma_index+1)
			
                #Add last dimension and get rid of the closing bracket 
                dimensions.append(int(strip_line[previous_coma_index+1:-1]))

                element_marginal_array = np.ndarray(shape=dimensions)

            if strip_line[0]=='#':
                if first_dim_line:
                    dimensions_names = []
                    if len(dimensions) > 1:
                        comma_index = strip_line.find(',')
                        opening_bracket_index = strip_line.find('[')
                        while opening_bracket_index != -1:
                            dimensions_names.append(strip_line[opening_bracket_index+1:comma_index])
                            opening_bracket_index = strip_line.find('[',comma_index) 
                            comma_index = strip_line.find(',',opening_bracket_index)
                    first_dim_line = False
                    dimensions_names.append(element_name)
                    dimension_names_dict[element_name] = dimensions_names
                    
                
                #update indices
                indices_array = []
                if len(dimensions) > 1:
                    comma_index = strip_line.find(',')
                    closing_brack_index = strip_line.find(']')					
                    while closing_brack_index != -1:
                        indices_array.append(int(strip_line[comma_index+1:closing_brack_index]))
                        opening_bracket_index = strip_line.find('[',closing_brack_index) 
                        comma_index = strip_line.find(',',opening_bracket_index)
                        closing_brack_index = strip_line.find(']',closing_brack_index+1)
				

            if strip_line[0]=='%':
                #read doubles
                coma_index = strip_line.find(',')
                marginals_values = []

                #Get rid of the %
                previous_coma_index = 0
                while coma_index != -1:
                    marginals_values.append(float(strip_line[previous_coma_index+1:coma_index]))
                    previous_coma_index = coma_index
                    coma_index = strip_line.find(',',coma_index+1)
			
                #Add last dimension and get rid of the closing bracket 
                marginals_values.append(float(strip_line[previous_coma_index+1:]))
                if len(marginals_values)!=dimensions[-1]:
                    print("problem")
                element_marginal_array[tuple(indices_array)] = marginals_values
        model_dict[element_name] = element_marginal_array				
        
        
    return [model_dict,dimension_names_dict]


"""
tcrdist3: Functional Programming Approach For Higher Memory Demand Use Cases
"""

import pwseqdist as pw
import pandas as pd
import numpy as np
from scipy import sparse

import os
from tcrdist import memory
import secrets
import parmap
import shutil
import multiprocessing

__all__ = ['_pw',
           '_pws',
           'compute_pw_sparse_out_of_memory',
           'compute_pws_sparse',
           'pw2dense']

def pw2dense(pw, maxd):
    """Make a pairwise distance matrix dense
    assuming -1 is used to encode D = 0"""
    pw = np.asarray(pw.todense())
    pw[pw == 0] = maxd + 1
    # pw[np.diag_indices_from(pw)] = 0
    pw[pw == -1] = 0
    return pw


def _pws(df, metrics, weights, kargs, df2 = None, cpu = 1, uniquify = True, store = False):
    """    
    _pws performs pairwise distance calculation across a multiple 
    columns of a Pandas DataFrame. This naturally permits calculation of
    a CDR-weighted tcrdistance that incorporates dissimilarity across 
    multiple complimentarity determining regions (see example below):

    Parameters 
    ----------
    df : pd.DataFrame
        Clones DataFrame containing, at a minimum, columns with CDR sequences
    df2 : pd.DataFrame or None
        Second clones DataFrame containing, at a minimum, columns with CDR sequences
    metrics : dict
        Dictionary of functions, specifying the distance metrics to apply to each CDR
    weights : dict
        Weights determining the contributions of each CDR distance to the aggregate distance
    kargs : dict
        Dictionary of Dictionaries
    cpu : int 
        Number of available cpus
    use_numba : bool
        If True, use must use a numba compatible metric 
    store : bool
        If False, only full tcrdist is returned. If True, 
        all component distance matrices are returned.
        
    Returns
    -------
    s : dictionary with tcr_distance.

    Example
    -------
    import pwseqdist as pw
    import pandas as pd
    from tcrdist.rep_funcs import _pw, _pw2
    
    # Define metrics for each region
    metrics = { "cdr3_a_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist}

    # Define weights
    weights = { 
                "cdr3_a_aa" : 3,
                "pmhc_a_aa" : 1,
                "cdr2_a_aa" : 1,
                "cdr1_a_aa" : 1,
                "cdr3_b_aa" : 3,
                "pmhc_b_aa" : 1,
                "cdr2_b_aa" : 1,
                "cdr1_b_aa" : 1}

    kargs = {   "cdr3_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr3_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}

    df = pd.DataFrame("dash2.csv")
    _pws(df = df, metrics = metrics, weights= weights, kargs=kargs, cpu = 1, store = False)
    """
    metric_keys = list(metrics.keys())
    weight_keys = list(weights.keys())
    assert metric_keys == weight_keys, "metrics and weights keys must be identical"
    
    if kargs is not None:
        kargs_keys  = list(kargs.keys())
        assert metric_keys == kargs_keys,  "metrics and kargs keys must be identical"
    
    tcrdist = None
    s = dict()
    for k in metric_keys:
        if df2 is None:
            pw_mat = _pw(seqs1 = df[k].values, metric = metrics[k], ncpus = cpu, uniqify= uniquify, **kargs[k])
        else:
            pw_mat = _pw(seqs1 = df[k].values, seqs2 = df2[k].values, metric = metrics[k], ncpus = cpu, uniqify= uniquify, **kargs[k])
            
        if store:
           s[k] = pw_mat 
        if tcrdist is None:
            tcrdist = np.zeros(pw_mat.shape, dtype=np.int16)
        tcrdist = tcrdist + (weights[k] * pw_mat)
    
    s['tcrdist'] = tcrdist
    return s


def _pw(metric, seqs1, seqs2=None, ncpus=1, uniqify= True, use_numba = False, **kwargs):
    """
    This is a wrapper for accessing pwseqdist version > 0.2.
    No matter what, it returns squareform results
    """
    pw_mat = pw.apply_pairwise_rect(metric = metric, 
                                    seqs1  = seqs1, 
                                    seqs2  = seqs2, 
                                    ncpus  = ncpus, 
                                    uniqify = uniqify, 
                                    use_numba = use_numba,
                                    **kwargs)
    # if len(pw_mat.shape) == 1:
    #     from scipy.spatial.distance import squareform
    #     pw_mat = squareform(pw_mat)

    return pw_mat


def compute_pws_sparse(df, metrics, weights, kargs, radius=50, df2=None, cpu=1, chunk_size=500, store=False, pm_pbar=True):
    """   
    compute_pw_sparse performs pairwise distance calculation across multiple 
    columns of a Pandas DataFrame. This naturally permits calculation of
    a CDR-weighted tcrdistance that incorporates dissimilarity across 
    multiple complimentarity determining regions (see example below):

    Radius is applied per chain.
    TODO: allow for different radius per chain

    Parameters 
    ----------
    df : pd.DataFrame
        Clones DataFrame containing, at a minimum, columns with CDR sequences
    df2 : pd.DataFrame or None
        Second clones DataFrame containing, at a minimum, columns with CDR sequences
    metrics : dict
        Dictionary of functions, specifying the distance metrics to apply to each CDR
    weights : dict
        Weights determining the contributions of each CDR distance to the aggregate distance
    kargs : dict
        Dictionary of Dictionaries
    cpu : int 
        Number of available cpus
    radius : int
        Only distances <= radius will be retained.
    chunk_size : int
        Number of rows that will make up each row chunk of [df x df2]
        which will be computed densely.
    store : bool
        IGNORED
        
    Returns
    -------
    s : dictionary with tcr_distance.

    Example
    -------
    import pwseqdist as pw
    import pandas as pd
    from tcrdist.rep_funcs import _pw, _pw2
    
    # Define metrics for each region
    metrics = { "cdr3_a_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_a_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr3_b_aa" : pw.metrics.nb_vector_tcrdist,
                "pmhc_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr2_b_aa" : pw.metrics.nb_vector_tcrdist,
                "cdr1_b_aa" : pw.metrics.nb_vector_tcrdist}

    # Define weights
    weights = { 
                "cdr3_a_aa" : 3,
                "pmhc_a_aa" : 1,
                "cdr2_a_aa" : 1,
                "cdr1_a_aa" : 1,
                "cdr3_b_aa" : 3,
                "pmhc_b_aa" : 1,
                "cdr2_b_aa" : 1,
                "cdr1_b_aa" : 1}

    kargs = {   "cdr3_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_a_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr3_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':3, 'ctrim':2, 'fixed_gappos':False},
                "pmhc_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr2_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True},
                "cdr1_b_aa" : {'use_numba': True, 'distance_matrix': pw.matrices.tcr_nb_distance_matrix, 'dist_weight': 1, 'gap_penalty':4, 'ntrim':0, 'ctrim':0, 'fixed_gappos':True}}

    df = pd.DataFrame("dash2.csv")
    _pws(df = df, metrics = metrics, weights= weights, kargs=kargs, cpu = 1, store = False)
    """
    metric_keys = [k for k in metrics.keys() if not 'cdr3' in k]
    weight_keys = [k for k in weights.keys() if not 'cdr3' in k]
    assert metric_keys == weight_keys, "metrics and weights keys must be identical"
    
    if kargs is not None:
        kargs_keys  = [k for k in kargs.keys() if not 'cdr3' in k]
        assert metric_keys == kargs_keys,  "metrics and kargs keys must be identical"
    
    n1 = df.shape[0]
    
    """Compute all but CDR3 as normal, but do not reexpand.
    Computing unique distances should not be memory or CPU intensive"""
    tcrdist = None
    components = dict()
    for k in metric_keys:
        if df2 is None:
            seqs2 = None
        else:
            seqs2 = df2[k].values

        """With reexapnd = False, returns: pw_mat, uind_i1, uind_i2"""
        pwmat, ind1, ind2 = pw.apply_pairwise_rect(metric=metrics[k], 
                                                    seqs1=df[k].values, 
                                                    seqs2=seqs2, 
                                                    ncpus=min(cpu, 2),
                                                    uniqify=True, 
                                                    reexpand=False,
                                                    **kargs[k])

        components[k] = (pwmat * weights[k], ind1, ind2) 

    """Can't do this because it will be huge. Also can't compute list of
    potential D < radius because that also could be too large.
    But need to pre-compute these non-CDR3 because otherwise the CDR3
    matrix won't be sparse enough with the radius.

    Solution: chunk the computation here. Only compute subsets of potential
    seqs pairs sparsely and spread across processors."""

    if cpu > 1 and n1 > chunk_size:
        """Chunk along df (rows) only"""
        chunk_func = lambda l, n: [np.array(l[i:i + n], dtype=np.int64) for i in range(0, len(l), n)]
        # chunksz = max(len(pw_indices) // cpu, 1)
        """Chunked indices is a list of arrays of indices"""
        """List of the chunked [chunk_size,] arrays"""
        chunked_indices = chunk_func(np.arange(n1, dtype=np.int64), chunk_size)
        
        with multiprocessing.Pool(cpu) as pool:
            dists = parmap.map(memory._sparse_cdr3_tcrdist_shard,
                               chunked_indices,
                               components,
                               df,
                               metrics,
                               weights,
                               kargs,
                               radius,
                               df2,
                               pm_parallel=True,
                               pm_pool=pool,
                               pm_pbar=pm_pbar)

        full_S = sparse.vstack(dists)
    else:
        full_S = memory._sparse_cdr3_tcrdist_shard(np.arange(n1, dtype=np.int64),
                                                components,
                                                df,
                                                metrics,
                                                weights,
                                                kargs,
                                                radius,
                                                df2)  
    
    return {'tcrdist': full_S}



def compute_n_tally_out_of_memory(fragments, 
                                 matrix_name = "rw_beta", 
                                 to_file = False, 
                                 to_memory = True,
                                 pm_processes = 2,
                                 **kwargs):
    """
    Parameters 
    ---------
    fragments : tuple 
        3-part tuple
            0 : TCRrep instance
            1 : list of rows in each fragment
            2 : filename holding the distances as a .npz
    matrix_name : str
        For 'beta' chain dists use 'rw_beta'
    to_file : bool
        If True, than a file is written to disk containing all the n_tally
    to_memory : bool
        If True, the n_tally is loaded directly to memory 
    **kwargs
        Keyword arguments are passed directly to tcridist.memory.gen_n_tally_on_fragment,
        including:  
            x_cols : list
                categorical variable 'epitope']  
            count_col : str
                column for counts  
            knn_neighbors : bool
                option to choose a fixed number of neighbors (NOT RECOMMENDED) 
            knn_radius : int
                maximum radius for finding neighbors

    Notes
    -----
    We envision that user may want to run nndif 
    multiple times with different categorical variables
    and may not want to go through the computational 
    intensive steps of computing TCRdistance each time. 

    This can be accomodated via setting cleanup to False:

    compute_pw_sparse_out_of_memory(cleanup=False) 

    Example 
    -------
    from scipy import sparse
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory
    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df,               
                organism = 'mouse',
                chains = ['beta'],
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = True,
                store_all_cdr = False)

    S, chunks = compute_pw_sparse_out_of_memory(tr, matrix_name = "rw_beta", max_distance = 1000, cleanup = False)
    # dest contains all the shards
    """

    # [(<tcrdist.repertoire.TCRrep at 0x14035b6d0>,
    #   range(0, 500),
    #   'd3be945e8956/0.rw_beta.npz'),

    # rearrange fragments in order (tr, ind, .npz, .csv)
    fragments =  [(x[0], x[1], x[2], f"{x[2]}.nndif.csv") for x in fragments ] 
    #fragments = [(tr, f"{dest}/{i}.{matrix_name}.npz", ind, f"{dest}/{i}.nndif.cvs") for i,ind in enumerate(row_chunks)] 
    csvfragments = parmap.starmap(memory.gen_n_tally_on_fragment, fragments, **kwargs, pm_pbar=True, pm_processes = pm_processes)
    if to_file:
        dest =os.path.dirname(csvfragments[0])
        nndiff_file = memory._concat_to_file(dest =dest, fragments =csvfragments)
        return nndiff_file

    if to_memory:
        nndiff = memory._concat_to_memory(fragments = csvfragments)
        return nndiff



def compute_pw_sparse_out_of_memory(tr,
                                    row_size      = 500,
                                    pm_processes  = 2,
                                    pm_pbar       = True,
                                    max_distance  = 50,
                                    matrix_name   = 'rw_beta',
                                    reassemble    = True,
                                    cleanup       = True):
    """
    Instead of calling TCRrep.compute_distances(), this 
    function permits a parallelizable approach that does 
    not require holding a large matrix in memory. 

    Default behavior is to reassemble a scipy
    sparse matrix from a set of sub matrices written to disk fragment. 
    With <reassemble = True>  function returns a scipy sparse matrix. 
    Space savings are achieved because any value above <max_distance> is set to zero. 
    True zero distances are set to 1. 

    Can be used to form a network of TCRs with tcrdistances < max_distance,

    Parameters
    ----------
    tr : TCRrep
        TCRrep instance with clone_df
    row_size : int
        How many rows to process in memory at once
    pm_processes : int
        Numbe of concurrent parallel processes to run at once
    pm_bar : bool 
        If True, show progress bar.
    max_distance : int
        Max distance
    matrix_name : str
        Name of matrix to return (i.e, 'rw_beta' or 'rw_alpha')
    reassemble: True
        If true, makes one matrix from all the sparse sub matrices. 
    cleanup: bool,
        if True, deletes temporary files. 

    Returns
    -------
    csr_full : sparse scipy matrix
            
    dest : str
        name of the folder that holds fragments

    
    Examples
    --------
    import numpy as np
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory

    df = pd.read_csv("dash.csv")
                          #(1)
    tr = TCRrep(cell_df = df,               #(2)
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = True,
                store_all_cdr = False)

    S = compute_pw_sparse_out_of_memory(tr, matrix_name = "rw_beta", max_distance = 1000)
    # S is a <1920x1920 sparse matrix of type '<class 'numpy.int16'>'
    M = S.todense()
    M[M==1] = 0 
    np.all(M == tr.pw_beta)
    S, chunks = compute_pw_sparse_out_of_memory(tr, matrix_name = "rw_beta", max_distance = 50)
    print(S)
    # S is a <1920x1920 sparse matrix of type '<class 'numpy.int16'>'
    """ 
    dest = secrets.token_hex(6)
    os.mkdir(dest)
    print(f"CREATED /{dest}/ FOR HOLDING DISTANCE OUT OF MEMORY")
    row_chunks = memory._partition(range(tr.clone_df.shape[0]), row_size)

    smatrix_chunks = [(tr, ind,  f"{dest}/{i}.{matrix_name}.npz") for i,ind in enumerate(row_chunks)]
    csrfragments = parmap.starmap(memory.gen_sparse_rw_on_fragment, 
            smatrix_chunks, 
            matrix_name = matrix_name, 
            max_distance=max_distance, 
            pm_pbar=pm_pbar, 
            pm_processes = pm_processes)
    if reassemble:
        csr_full = memory.collapse_csrs([x[2] for x in smatrix_chunks])
        print(f"RETURNING scipy.sparse csr_matrix w/dims {csr_full.shape}")
    else: 
        csr_full = None
        
    if cleanup: 
        assert os.path.isdir(dest)
        print(f"CLEANING UP {dest}")
        shutil.rmtree(dest)
    
    
    return csr_full, smatrix_chunks



def compute_pw_sparse_out_of_memory2(tr,
                                    row_size      = 500,
                                    pm_processes  = 2,
                                    pm_pbar       = True,
                                    max_distance  = 50,
                                    reassemble    = True,
                                    cleanup       = True,
                                    assign        = True):
    """
    Instead of calling TCRrep.compute_distances(), this 
    function permits a parallelizable approach that does 
    not require holding a large matrix in memory. 

    Default behavior is to reassemble a scipy
    sparse matrix from a set of sub matrices written to disk fragment. 
    With <reassemble = True>  function returns a scipy sparse matrix. 
    Space savings are achieved because any value above <max_distance> is set to zero. 
    True zero distances are set to -1. 

    Can be used to form a network of TCRs with tcrdistances < max_distance,

    Parameters
    ----------
    tr : TCRrep
        TCRrep instance with clone_df
    row_size : int
        How many rows to process in memory at once
    pm_processes : int
        Numbe of concurrent parallel processes to run at once
    pm_bar : bool 
        If True, show progress bar.
    max_distance : int
        Max distance
    matrix_name : str
        Name of matrix to return (i.e, 'rw_beta' or 'rw_alpha')
    reassemble: True
        If true, makes one matrix from all the sparse sub matrices. 
    cleanup: bool,
        if True, deletes temporary files. 
    assign : bool 
        if True, assigns pw sparse matrices to TCRrep object.
        That is TCRrep.pw_beta, TCRrep.pw_alpha will be assigned 
        the reassembled spare matrces.

    Returns
    -------
    csr_full : sparse scipy matrix
            
    dest : str
        name of the folder that holds fragments

    
    Examples
    --------
    import numpy as np
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory

    df = pd.read_csv("dash.csv")
                          #(1)
    tr = TCRrep(cell_df = df,               #(2)
                organism = 'mouse', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = True,
                store_all_cdr = False)

    S = compute_pw_sparse_out_of_memory(tr, matrix_name = "rw_beta", max_distance = 1000)
    # S is a <1920x1920 sparse matrix of type '<class 'numpy.int16'>'
    M = S.todense()
    M[M==1] = 0 
    np.all(M == tr.pw_beta)
    S, chunks = compute_pw_sparse_out_of_memory(tr, matrix_name = "rw_beta", max_distance = 50)
    print(S)
    # S is a <1920x1920 sparse matrix of type '<class 'numpy.int16'>'
    """ 

    # Early warning to save heartache
    if assign is True and reassemble is False:
        raise ValueError("If you want to assign results to a TCRrep instance, you must set reassemble to True")

    dest = secrets.token_hex(6)
    os.mkdir(dest)
    print(f"CREATED /{dest}/ FOR HOLDING DISTANCE OUT OF MEMORY")
    row_chunks = memory._partition(range(tr.clone_df.shape[0]), row_size)


    smatrix_chunks = [(tr, ind,  f"{dest}/{i}") for i,ind in enumerate(row_chunks)]
    csrfragments = parmap.starmap(memory.gen_sparse_rw_on_fragment2, 
            smatrix_chunks,  
            max_distance=max_distance, 
            pm_pbar=pm_pbar, 
            pm_processes = pm_processes)

    if reassemble:
        csr_full_dict = dict()
        for chain in tr.chains:
            chain_str = f"rw_{chain}"
            csr_full = memory.collapse_csrs([f"{x[2]}.{chain_str}.npz" for x in smatrix_chunks])
            print(f"RETURNING scipy.sparse csr_matrix w/dims {csr_full.shape}")
            csr_full_dict[chain] = csr_full
    else: 
        csr_full_dict= None
       
    if assign:
        for chain in tr.chains:
            setattr(tr, f"pw_{chain}", csr_full_dict[chain])

    if cleanup: 
        assert os.path.isdir(dest)
        print(f"CLEANING UP {dest}")
        shutil.rmtree(dest)
    
    return csr_full_dict, smatrix_chunks




def compute_n_tally_out_of_memory2(fragments, 
                                 to_file = False, 
                                 to_memory = True,
                                 pm_processes = 2,
                                 **kwargs):
    """
    Parameters 
    ---------
    fragments : tuple 
        3-part tuple
            0 : TCRrep instance
            1 : list of rows in each fragment
            2 : filename holding the distances as a .npz
    matrix_name : str
        For 'beta' chain dists use 'rw_beta'
    to_file : bool
        If True, than a file is written to disk containing all the n_tally
    to_memory : bool
        If True, the n_tally is loaded directly to memory 
    **kwargs
        Keyword arguments are passed directly to tcridist.memory.gen_n_tally_on_fragment,
        including:  
            x_cols : list
                categorical variable 'epitope']  
            count_col : str
                column for counts  
            knn_neighbors : bool
                option to choose a fixed number of neighbors (NOT RECOMMENDED) 
            knn_radius : int
                maximum radius for finding neighbors

    Notes
    -----
    We envision that user may want to run nndif 
    multiple times with different categorical variables
    and may not want to go through the computational 
    intensive steps of computing TCRdistance each time. 

    This can be accomodated via setting cleanup to False:

    compute_pw_sparse_out_of_memory(cleanup=False) 

    Example 
    -------
    from scipy import sparse
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_funcs import  compute_pw_sparse_out_of_memory
    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df,               
                organism = 'mouse',
                chains = ['beta'],
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = True,
                store_all_cdr = False)

    S, chunks = compute_pw_sparse_out_of_memory(tr, matrix_name = "rw_beta", max_distance = 1000, cleanup = False)
    # dest contains all the shards
    """

    # [(<tcrdist.repertoire.TCRrep at 0x14035b6d0>,
    #   range(0, 500),
    #   'd3be945e8956/0.rw_beta.npz'),

    # rearrange fragments in order (tr, ind, .npz, .csv)
    fragments =  [(x[0], x[1], x[2], f"{x[2]}.nndif.csv") for x in fragments ] 
    #fragments = [(tr, f"{dest}/{i}.{matrix_name}.npz", ind, f"{dest}/{i}.nndif.cvs") for i,ind in enumerate(row_chunks)] 
    csvfragments = parmap.starmap(memory.gen_n_tally_on_fragment2, fragments, **kwargs, pm_pbar=True, pm_processes = pm_processes)
    if to_file:
        dest =os.path.dirname(csvfragments[0])
        nndiff_file = memory._concat_to_file(dest =dest, fragments =csvfragments)
        return nndiff_file

    if to_memory:
        nndiff = memory._concat_to_memory(fragments = csvfragments)
        return nndiff






import pytest 



def test_introduction_2():
    import pandas as pd
    from tcrdist.repertoire import TCRrep
    from tcrdist.adpt_funcs import import_adaptive_file, adaptive_to_imgt
    
    df = import_adaptive_file(adaptive_filename = 'Adaptive2020.tsv')
    # For larger datasets, make sure compute_distances is set to False, 
    # see: https://tcrdist3.readthedocs.io/en/latest/bulkdata.html
    tr = TCRrep(cell_df = df, 
                organism = 'human', 
                chains = ['beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv', compute_distances = False)

    """Lookup *01 IMGT allele corresponding with an Adaptive gene name"""
    assert adaptive_to_imgt['human']['TCRBV30'] == 'TRBV30*01'    



import os
import re
import pandas as pd
import numpy as np
import warnings
from tcrsampler.sampler import TCRsampler

def _default_sampler(organism = 'human', chain = 'beta'):
    assert organism in ['human', 'mouse']
    assert chain in ['beta','alpha']

    default_tcrsampler_generator = {
        ('human','beta'): 
            _default_tcrsampler_human_beta,
        ('human','alpha'): 
            _default_tcrsampler_human_alpha,
        ('mouse','beta'): 
            _default_tcrsampler_mouse_beta,
        ('mouse','alpha'): 
            _default_tcrsampler_mouse_alpha, 
        }[(organism, chain)]
    
    return default_tcrsampler_generator

def _default_sampler_olga(organism = 'human', chain = 'beta'):
    assert organism in ['human', 'mouse']
    assert chain in ['beta','alpha']
    if organism == "mouse" and chain == "alpha":
        raise KeyError("No currenlty available mouse-alpha chain background form olga")

    default_tcrsampler_generator = {
        ('human','beta'): 
            _default_tcrsampler_olga_human_beta,
        ('human','alpha'): 
            _default_tcrsampler_olga_human_alpha,
        ('mouse','beta'): 
            _default_tcrsampler_olga_mouse_beta
        }[(organism, chain)]
    
    return default_tcrsampler_generator


def _default_tcrsampler_olga_human_beta(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background =  'olga_human_beta_t.sampler.tsv'
        
    if default_background_if_missing is None:
        default_background_if_missing ='olga_sampler.zip'
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_olga_human_alpha(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background =  'olga_human_alpha_t.sampler.tsv'
        
    if default_background_if_missing is None:
        default_background_if_missing ='olga_sampler.zip'
    
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_olga_mouse_beta(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background =  'olga_mouse_beta_t.sampler.tsv'
        
    if default_background_if_missing is None:
        default_background_if_missing ='olga_sampler.zip'
    
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t


def _default_tcrsampler_human_beta(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background = 'britanova_human_beta_t_cb.tsv.sampler.tsv'
        
    if default_background_if_missing is None:
        default_background_if_missing ='britanova_human_beta_t_cb.tsv.sampler.tsv.zip'
    
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_human_alpha(default_background = None, default_background_if_missing=None ):
    """
    Responsible for providing the default human alpha sampler 'ruggiero_human_alpha_t.tsv.sampler.tsv'
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background = 'ruggiero_human_alpha_t.tsv.sampler.tsv'
    if default_background_if_missing is None:
        default_background_if_missing =  'ruggiero_human_alpha_t.tsv.sampler.tsv.zip'
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_mouse_beta(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default mouse beta sampler

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler

    if default_background is None:
        default_background = 'ruggiero_mouse_beta_t.tsv.sampler.tsv'
    if default_background_if_missing is None:
        default_background_if_missing =  'ruggiero_mouse_sampler.zip'

    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_mouse_alpha(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default mouse alpha sampler
    """
    from tcrsampler.sampler import TCRsampler
    
    if default_background is None:
        default_background = 'ruggiero_mouse_alpha_t.tsv.sampler.tsv'
    if default_background_if_missing is None:
        default_background_if_missing =  'ruggiero_mouse_sampler.zip'
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def allele_01(genename):
    """
    >>> allele_01('TRBV19*02')
    'TRBV19*01'
    """
    g,a = re.match(pattern = '(.*)([0-9])$', string= genename).groups()
    return(f"{g}1")


from tcrdist.regex import _index_to_regex_str, _index_to_seqs, _matrix_to_regex, _index_to_matrix, _list_to_regex_component
import pytest
import pandas as pd
import os
from tcrdist.repertoire import TCRrep	
from tcrdist.regex import _index_to_matrix

"""
Fixture
"""
fn = os.path.join('dash.csv')
df = pd.read_csv(fn , sep = ",")
df = df[(df.epitope == 'NP')].reset_index()
tr = TCRrep( cell_df = df,
        organism = 'mouse',
        chains = ['beta'],
        db_file = 'alphabeta_gammadelta_db.tsv')
"""
unit and coverage tests for the /tcrdist/regex.py 
"""
def test_list_to_regex_component_sinlge():
	assert _list_to_regex_component(['A'],max_ambiguity = 3) == 'A'

def test_list_to_regex_component_group():
	assert _list_to_regex_component(['A','T','C'],max_ambiguity = 3) == '[ATC]'

def test_list_to_regex_component_group_more_than_max_ambiguity():
	assert _list_to_regex_component(['A','T','C','G'],max_ambiguity = 3) == '.'

def test_list_to_regex_component_change_ambiguity_threshold():
	assert _list_to_regex_component(['A','T','C','G'],max_ambiguity = 4) == '[ATCG]'

def test_list_to_regex_component_gap_makes_group_optional():
	assert _list_to_regex_component(['A','-','C']) == '[AC]?'

def test__matrix_to_regex():
	import numpy as np
	import pandas as pd
	arr = np.array([   [ 0. ,  3.3,  0. ,  0. ,  0.4,  0. ,  0. ,  0. ],
					   [ 0. ,  0. ,  0. ,  0. ,  0.  ,  0. ,  3.6,  4.5],
					   [ 0. ,  0. ,  0. ,  0. ,  0.  ,  0. ,  0. ,  0. ],
					   [ 0. ,  0. ,  0. ,  0. ,  0.  ,  4.5,  0. ,  0. ],
					   [ 4.5,  0. ,  0.4,  0. ,  0.  ,  0. ,  0. ,  0. ],
					   [ 0. ,  0. ,  0. ,  0. ,  2.67,  0. ,  0. ,  0. ],
					   [ 0. ,  0. ,  0. ,  0. ,  0.  ,  0. ,  0. ,  0. ],
					   [ 0. ,  0. ,  0. ,  0.1,  0.  ,  0. ,  0. ,  0. ]])
	matrix = pd.DataFrame(arr, 
		index = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G'], 
		columns = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G'])
	assert _matrix_to_regex(matrix) == '(G[AQ]D)'

def test_index_to_matrix(tr =tr ):
	r = _index_to_matrix(ind = [1,2,3], clone_df=tr.clone_df, pwmat = tr.pw_beta, col = 'cdr3_b_aa', centroid = None )
	assert isinstance(r, pd.DataFrame)
	assert r.shape == (24,14)
	assert r.columns.to_list() == ['C', 'A', 'S', 'S', 'G', 'K', 'Q', 'N', 'T', 'S', 'Q', 'F', 'T', 'F']

def test_index_to_seqs(ind = [1,2,3], clone_df = tr.clone_df, col = 'cdr3_b_aa'):
	dfnode   = clone_df.iloc[ind,].copy()
	seqs = dfnode[col].to_list()
	assert seqs == ['CASSGGANTGQLYF', 'CASSGKQNTSQFTF', 'CASSLKQGRSDYTF']

def test_index_to_regex_str():
	"""
	ind : list 
		iloc row index in clone_df
	clone_df : pd.DataFrae
		DataFrame with cdr3 sequence
	pwmat: np.ndarray
		Matrix with pairwise inofrmation
	col : str
		'cdr3_b_aa', 
	ntrim : int
		default 3, 
	ctrim : int
		default 2,  
	max_ambiguity : int
		default 3, the maximum number of amino acids at a position before a '.' wildcard is applied
	"""
	regex_pattern = _index_to_regex_str(ind = [1,2,3], 
										clone_df = tr.clone_df, 
										pwmat = tr.pw_beta, 
										centroid = None,
										col = 'cdr3_b_aa', 
										ntrim = 3, 
										ctrim = 2,  
										max_ambiguity = 3)

	assert regex_pattern == '(S[GL][GK][AQ][NG][RT][GS][DQ][LFY])'




import pytest
import numpy as np
from tcrdist import repertoire_db

def test_extract_cdrs_without_gaps():
    """ specific case """
    a = repertoire_db._extract_cdrs_without_gaps('GQGVEQ.P.AKLMSVEGTFARVNCTYSTSG......FNGLSWYQQREGQAPVFLSYVVL....DGLKDS.....GHFSTFLSRSN.GYSYLLLTELQIKDSASYLCAVR..',
    '28-39;57-66;82-88;106-111')
    b = ['TSGFNG', 'VVLDGL', 'SRSNGY', 'CAVR']
    
    assert a==b

def test_extract_aligned_cdrs():
    """ specific case """
    a = repertoire_db._extract_aligned_cdrs('GQGVEQ.P.AKLMSVEGTFARVNCTYSTSG......FNGLSWYQQREGQAPVFLSYVVL....DGLKDS.....GHFSTFLSRSN.GYSYLLLTELQIKDSASYLCAVR..',
    '28-39;57-66;82-88;106-111')
    b = ['TSG......FNG', 'VVL....DGL', 'SRSN.GY', 'CAVR..']
    
    assert a==b




import pytest 



def test_introduction_1():
    import pandas as pd
    from tcrdist.repertoire import TCRrep

    df = pd.read_csv("dash.csv")
    tr = TCRrep(cell_df = df, 
                organism = 'mouse', 
                chains = ['alpha','beta'], 
                db_file = 'alphabeta_gammadelta_db.tsv',
                compute_distances = False)


import pandas as pd
import numpy as np

import hierdiff as hd

__all__ = ['neighborhood_diff',
           'hcluster_diff',
           'member_summ']

def neighborhood_diff(clone_df, pwmat, x_cols, count_col='count', knn_neighbors=None, knn_radius=None, subset_ind=None, cluster_ind=None, test_method='fishers'):
    """Tests for association of categorical variables in x_cols with the neighborhood
    around each TCR in clone_df. The neighborhood is defined by the K closest neighbors
    using pairwise distances in pwmat, or defined by a distance radius, knn_radius.

    Uses hierdiff package (available on PyPI) for tallying counts in each cluster
    and running tests.

    The statistical tests made available by this function are limited and meant only
    as a way to scan for signals. More sophisticated testing/modeling frameworks
    should be considered for real-world problems.

    Use test_method = None to return a table of counts for all neighborhoods that can be saved as
    a CSV and used to run other, more sophisticated tests (e.g. edgeR or other regressions).

    Use Fisher's exact test (test_method='fishers') to detect enrichment/association of the neighborhood
    with one binary variable. For example, test the 2 x 2 table for each clone:

    +----+----+-------+--------+
    |         |  Neighborhood  |
    |         +-------+--------+
    |         | MEM+  |   MEM- |
    +----+----+-------+--------+
    |VAR |  0 | a     |    b   |
    |    +----+-------+--------+
    |    |  1 | c     |    d   |
    +----+----+-------+--------+

    Use the chi-squared test (test='chi2') to detect association across multiple variables.
    Note that with sparse neighborhoods Chi-squared tests are unreliable.

    Use the Cochran-Mantel-Haenszel test (test='chm') to test stratified 2 x 2 tables:
    one VAR vs. neighborhood, over several strata defined in other variables.
    Use x_cols[0] as the primary (binary) variable and other x_cols for the categorical
    strata-defining variables. This tests the overall null that OR = 1 for x_cols[0].
    A test is also performed for homogeneity of the ORs among the strata (Breslow-Day test).

    Params
    ------
    clone_df : pd.DataFrame [nclones x metadata]
        Contains metadata for each clone.
    pwmat : np.ndarray [nclones x nclones]
        Square distance matrix for defining neighborhoods
    x_cols : list
        List of columns to be tested for association with the neighborhood. 
        (If test_method ='fishers' than x_cols must be length 1 and contain a binary variable)
    count_col : str
        Column in clone_df that specifies counts.
        Default none assumes count of 1 cell for each row.
    knn_neighbors : int
        Number of neighbors to include in the neighborhood.
    knn_radius : float
        Radius for inclusion of neighbors within the neighborhood.
        Specify K or R but not both.
    cluster_ind : None or np.ndarray
        Indices into df specifying the neighborhoods for testing.
    test_method : str or None
        Specifies Fisher's exact test ("fishers"), Chi-squared ("chi2") or
        Cochran-Mantel-Haenszel test ("chm") for testing.

    Returns
    -------
    res : pd.DataFrame [nclones x results]
        Results from testing the neighborhood around each clone."""
    res = hd.neighborhood_tally(df_pop=clone_df,
                                  pwmat=pwmat,
                                  x_cols=x_cols,
                                  count_col=count_col,
                                  knn_neighbors=knn_neighbors,
                                  knn_radius=knn_radius)
    if not test_method is None:
        res = hd.cluster_association_test(res, y_col='cmember', method=test_method)
    return res

def hcluster_diff(clone_df, pwmat, x_cols, Z=None, count_col='count', subset_ind=None, hclust_method='complete', optimal_ordering=True, test_method='fishers'):
    """Tests for association of categorical variables in x_cols with each cluster/node
    in a hierarchical clustering of TCR clones with distances in pwmat.

    Uses hierdiff package (available on PyPI) for tallying counts in each cluster
    and running tests.

    The statistical tests made available by this function are limited and meant only
    as a way to scan for signals. More sophisticated testing/modeling frameworks
    should be considered for real-world problems.

    Use test_method = None to return a table of counts for all neighborhoods that can be saved as
    a CSV and used to run other, more sophisticated tests (e.g. edgeR or other regressions).

    Use Fisher's exact test (test='fishers') to detect enrichment/association of the neighborhood/cluster
    with one binary variable.

    Tests the 2 x 2 table for each clone:

    +----+----+-------+--------+
    |         |    Cluster     |
    |         +-------+--------+
    |         |  MEM+ |  MEM-  |
    +----+----+-------+--------+
    |VAR |  0 |   a   |    b   |
    |    +----+-------+--------+
    |    |  1 |   c   |    d   |
    +----+----+-------+--------+

    Use the chi-squared test (test='chi2') to detect association across multiple categorical variables.
    Note that with small clusters Chi-squared tests are unreliable.

    Use the Cochran-Mantel-Haenszel test (test='chm') to test stratified 2 x 2 tables:
    one VAR vs. neighborhood, over several strata defined in other variables.
    Use x_cols[0] as the primary (binary) variable and other x_cols for the categorical
    strata-defining variables. This tests the overall null that OR = 1 for x_cols[0].
    A test is also performed for homogeneity of the ORs among the strata (Breslow-Day test).

    Params
    ------
    clone_df : pd.DataFrame [nclones x metadata]
        Contains metadata for each clone.
    pwmat : np.ndarray [nclones x nclones]
        Square distance matrix for defining neighborhoods
    x_cols : list
        List of columns to be tested for association with the neighborhood
    count_col : str
        Column in clone_df that specifies counts.
        Default none assumes count of 1 cell for each row.
    subset_ind : None or np.ndarray with partial index of df, optional
        Provides option to tally counts only within a subset of df, but to maintain the clustering
        of all individuals. Allows for one clustering of pooled TCRs,
        but tallying/testing within a subset (e.g. participants or conditions)
    hclust_method : str
        Method for hierarchical clustering, passed to the scipy.clustering.hierarchy
        linkage function.
    optimal_ordering : bool
        Flag passed to the scipy.clustering.hierarchy linkage function to improve
        visual tree layout. Can be slow for large trees.
    test_method : str or None
        Specifies Fisher's exact test ("fishers"), Chi-squared ("chi2") or
        Cochran-Mantel-Haenszel test ("chm") for testing.

    Returns
    -------
    res : pd.DataFrame [nclusters x results]
        Results from testing each cluster.
    Z : linkage matrix [clusters, 4]
        Clustering result returned from scipy.cluster.hierarchy.linkage"""
    res, Z = hd.hcluster_tally(df=clone_df,
                                  pwmat=pwmat,
                                  x_cols=x_cols,
                                  Z=Z,
                                  count_col=count_col,
                                  subset_ind=subset_ind,
                                  method='complete',
                                  optimal_ordering=optimal_ordering)
    if not test_method is None:
        res = hd.cluster_association_test(res, y_col='cmember', method=test_method)
    return res, Z


def member_summ(res_df, clone_df, key_col = 'neighbors_i', count_col='count', addl_cols=[], addl_n=1):
    """Return additional summary info about each result (row)) based on the members of the cluster.

    This is helpful for preparing strings to add to the tooltip in hierdiff.plot_hclust_props.
    
    Parameters
    ----------
    res_df : pd.DataFrame [nclusters x result cols]
        Returned from neighborhood_diff or hcluster_diff
    clone_df : pd.DataFrame [nclones x metadata]
        Contains metadata for each clone.
    key_col : str
        Column in res_df that specifies the iloc of members in the clone_df
    count_col : str
        Column in clone_df that specifies counts.
        Default none assumes count of 1 cell for each row.
    addl_cols : list
        Columns to summarize
    addl_n : int
        Number of top N clones to include in the summary of
        each cluster.

    Returns
    -------
    summ : pd.DataFrame [nclusters x summary columns]
        Columns that can be joined with res_df


    Example
    -------
    summ_df = member_summ(res_df, clone_df)
    res_df = res_df.join(summ_df, how='left')"""
    def _top_N_str(m, col, count_col, N):
        gby = m.groupby(col)[count_col].agg('sum')
        gby = 100 * gby / gby.sum()
        gby = gby.sort_values(ascending=False)
        out = ', '.join(['%s (%2.1f%%)' % (idx, v) for idx, v in gby.items()][:N])
        return out
    
    split = []
    for resi, res_row in res_df.iterrows():
        m = clone_df.iloc[res_row[key_col]]

        mode_i = m[count_col].idxmax()
        summ = {}
        for c in [c for c in clone_df.columns if 'cdr3' in c]:
            summ[c] = _top_N_str(m, c, count_col, 1)
        for c in [c for c in clone_df.columns if 'gene' in c]:
            summ[c] = _top_N_str(m, c, count_col, 3)

        x_val_cols = [c for c in res_df.columns if 'x_val_' in c]
        x_freq_cols = [c for c in res_df.columns if 'x_freq_' in c]
        
        for label_col, freq_col in zip(x_val_cols, x_freq_cols):
            summ[res_row[label_col]] = np.round(res_row[freq_col], 3)

        for c in [c for c in addl_cols]:
            summ[c] = _top_N_str(m, c, count_col, addl_n)
        summ = pd.Series(summ, name=resi)
        split.append(summ)
    summ = pd.DataFrame(split)
    return summ



import os
import re
import pandas as pd
import numpy as np
import warnings
from progress.bar import IncrementalBar
from tcrdist.pgen import OlgaModel

class TCRtree():
    def __init__(self,tcrrep, html_name):
        """
        Automatic Hierarchical Cluster Plotting.

        This is a class that automates the process of producing 
        interactive TCR hierarchical clustering tree.

        These trees can be constructed manually see `docs
        <https://tcrdist3.readthedocs.io/en/latest/motif_gallery.html>_.
        <http://www.python.org/>`_.
        Parameters
        ----------

        html_name : str
            name for html file output e.g., 'DEFAULT.html'
        pwmat_str_b : str
            name of pairwise matrix attribute to be used for clustering beta chains e.g., 'pw_beta'
        pwmat_str_a : str
            name of pairwise matrix attribute to be used for clustering alpha chains e.g., 'pw_alpha'
        single : bool
            If true, make summary based on each clone being present in single-copy, 
            otherwise, the 'count' column is used when calculating percentages.
            NOTE: 'count_col' in default_hcluster_diff_kwargs can also set to 
            'single'. If true, diversity metrics will also be based on clones 
            rather than clonal abundances. 
        generate_svgs : bool 
            If True, SVG logos are produced for each node where .hcluster_df['prune'] is 0
        verbose : bool
            report on status


        default_hcluster_diff_kwargs: dict
            kwargs dictionary for (tcrdist.rep_diff.hcluster_diff)

                clone_df : pd.DataFrame [nclones x metadata]
                    Contains metadata for each clone.
                pwmat : np.ndarray [nclones x nclones]
                    Square distance matrix for defining neighborhoods
                x_cols : list
                    List of columns to be tested for association with the neighborhood
                count_col : str
                    Column in clone_df that specifies counts.
                    Default none assumes count of 1 cell for each row.
                subset_ind : None or np.ndarray with partial index of df, optional
                    Provides option to tally counts only within a subset of df, but to maintain the clustering
                    of all individuals. Allows for one clustering of pooled TCRs,
                    but tallying/testing within a subset (e.g. participants or conditions)
                hclust_method : str
                    Method for hierarchical clustering, passed to the scipy.clustering.hierarchy
                    linkage function.
                optimal_ordering : bool
                    Flag passed to the scipy.clustering.hierarchy linkage function to improve
                    visual tree layout. Can be slow for large trees.
                test_method : str or None
                    Specifies Fisher's exact test ("fishers"), Chi-squared ("chi2") or
                    Cochran-Mantel-Haenszel test ("chm") for testing.
        default_member_summ_kwargs: dict
            kwargs dictionary for (tcrdist.rep_diff.member_summ)

                Return additional summary info about each result (row)) based on the
                members of the cluster. This is helpful for preparing strings 
                to add to the tooltip in hierdiff.plot_hclust_props.

                res_df : pd.DataFrame [nclusters x result cols]
                    Returned from neighborhood_diff or hcluster_diff
                clone_df : pd.DataFrame [nclones x metadata]
                    Contains metadata for each clone.
                key_col : str
                    Column in res_df that specifies the iloc of members in the clone_df
                count_col : str
                    Column in clone_df that specifies counts.
                    Default none assumes count of 1 cell for each row.
                addl_cols : list
                    Columns to summarize
                addl_n : int
                    Number of top N clones to include in the summary of
                    each cluster.
        default_plot_hclust_props : dict
            kwargs dictionary for (hierdiff.plot_hclust_props)

                Plot tree of linkage-based hierarchical clustering, with nodes colored using stacked bars
                representing proportion of cluster members associated with specific conditions. Nodes also optionally
                annotated with pvalue, number of members or cluster ID.
                
                Z : linkage matrix
                    Result of calling sch.linkage on a compressed pair-wise distance matrix
                res : pd.DataFrame
                    Result from calling hcluster_diff, with observed/frequencies and p-values for each node
                alpha_col : str
                    Column in res to use for 'alpha' annotation
                alpha : float
                    Threshold for plotting the stacked bars and annotation
                colors : tuple of valid colors
                    Used for stacked bars of conditions at each node
                prune_col : str/column in res
                    Column of res that indicates whether a result/node can be pruned from the tree.
                    The tree will not print any pruned nodes that only contain other pruned nodes.
        """
        self.tcrrep = tcrrep
        self.html_name = html_name

        kws1,kws2,kws3 = _get_default_kwargs(chains = self.tcrrep.chains)
        self.default_hcluster_diff_kwargs = kws1
        self.default_member_summ_kwargs   = kws2
        self.default_plot_hclust_props    = kws3
        self.prune = 3
        self.combine_olga = True
        self.pwmat_str_b = 'pw_beta'
        self.pwmat_str_a = 'pw_alpha'
        self.single = True
        self.generate_svgs = True
        self.verbose = True

    def build_tree( self ):

        print(self.default_plot_hclust_props['tooltip_cols'])
        _auto_hdiff2(   tcrrep                       = self.tcrrep,
                        html_name                    = self.html_name,
                        pwmat_str_b                  = self.pwmat_str_b,
                        pwmat_str_a                  = self.pwmat_str_a,
                        single                       = self.single,
                        generate_svgs                = self.generate_svgs ,
                        combine_olga                 = self.combine_olga,
                        prune                        = self.prune,
                        verbose                      = self.verbose,
                        default_hcluster_diff_kwargs = self.default_hcluster_diff_kwargs,
                        default_member_summ_kwargs   = self.default_member_summ_kwargs,
                        default_plot_hclust_props    = self.default_plot_hclust_props)



def _get_default_kwargs(chains = ['beta']):
    default_hcluster_diff_kwargs = {
        'clone_df' : None, 
        'pwmat' : None, 
        'x_cols' : ['epitope'],#None,#['epitope'], 
        'Z' : None, 
        'count_col' : 'count', 
        'subset_ind' : None, 
        'hclust_method' : 'complete', 
        'optimal_ordering' : True, 
        'test_method' : 'fishers'
        }
    default_member_summ_kwargs = {
        'key_col' : 'neighbors_i', 
        'count_col' : 'count', 
        'addl_cols' :['subject'], 
        'addl_n' : 1
        }

    if len(chains) == 1:
        chain = chains[0]
        chain_abr = chain[0]
        default_plot_hclust_props = {
            'title':'', 
            'alpha_col':'pvalue', 
            'alpha':0.05, 
            'tooltip_cols':['subject',
                            'mean_dist',
                            'pct_dist_75',
                            'pct_dist_50',
                            'pct_dist_25',
                            'fuzzy_simpson_diversity_75', 
                            'fuzzy_simpson_diversity_50',
                            'fuzzy_simpson_diversity_25',
                            f'cdr3_{chain_abr}_aa', 
                            f'v_{chain_abr}_gene',
                            f'j_{chain_abr}_gene',
                            f'svg_{chain}',
                            f'svg_raw_{chain}',
                            f'ref_size_{chain}', 
                            f'ref_unique_{chain}',
                            f'percent_missing_{chain}']}
    elif len(chains) == 2:
        chain = chains[0]
        chain_abr = chain[0]
        chain2 = chains[1]
        chain2_abr = chain2[0]
        default_plot_hclust_props = {
            'title':'', 
            'alpha_col':'pvalue', 
            'alpha':0.05, 
            'tooltip_cols':[
                            'mean_dist',
                            'pct_dist_75',
                            'pct_dist_50',
                            'pct_dist_25',
                            'fuzzy_simpson_diversity_75', 
                            'fuzzy_simpson_diversity_50',
                            'fuzzy_simpson_diversity_25',
                            f'cdr3_{chain_abr}_aa', 
                            f'v_{chain_abr}_gene',
                            f'j_{chain_abr}_gene',
                            f'svg_{chain}',
                            f'svg_raw_{chain}',
                            f'ref_size_{chain}', 
                            f'ref_unique_{chain}',
                            f'percent_missing_{chain}',
                            f'cdr3_{chain2_abr}_aa', 
                            f'v_{chain2_abr}_gene',
                            f'j_{chain2_abr}_gene',
                            f'svg_{chain2}',
                            f'svg_raw_{chain2}',
                            f'ref_size_{chain2}', 
                            f'ref_unique_{chain2}',
                            f'percent_missing_{chain2}']}
    else:
        raise ValueError('Only 1 or 2 chains can be provided')

    return default_hcluster_diff_kwargs, default_member_summ_kwargs, default_plot_hclust_props


def _auto_hdiff2(tcrrep,
                html_name = 'DEFAULT.html',
                pwmat_str_b = 'pw_beta',
                pwmat_str_a = 'pw_alpha',
                single = True,
                generate_svgs = True,
                combine_olga = False,
                verbose = True,
                prune = 3,
                default_hcluster_diff_kwargs = _get_default_kwargs(chains = ['beta'])[0],
                default_member_summ_kwargs   = _get_default_kwargs(chains = ['beta'])[1],
                default_plot_hclust_props    = _get_default_kwargs(chains = ['beta'])[2]):
    """
    Automatic Hierarchical Cluster Plotting

    Parameters
    ----------

    html_name : str
        name for html file output e.g., 'DEFAULT.html'
    pwmat_str_b : str
        name of pairwise matrix attribute to be used for clustering beta chains e.g., 'pw_beta'
    pwmat_str_a : str
        name of pairwise matrix attribute to be used for clustering alpha chains e.g., 'pw_alpha'
    single : bool
        If true, make summary based on each clone being present in single-copy, 
        otherwise, the 'count' column is used when calculating percentages.
        NOTE: 'count_col' in default_hcluster_diff_kwargs can also set to 
        'single'. If true, diversity metrics will also be based on clones 
        rather than clonal abundances. 
    generate_svgs : bool 
        If True, SVG logos are produced for each node where .hcluster_df['prune'] is 0
    verbose : bool
        report on status
    default_hcluster_diff_kwargs: dict
        kwargs dictionary for (tcrdist.rep_diff.hcluster_diff)

            clone_df : pd.DataFrame [nclones x metadata]
                Contains metadata for each clone.
            pwmat : np.ndarray [nclones x nclones]
                Square distance matrix for defining neighborhoods
            x_cols : list
                List of columns to be tested for association with the neighborhood
            count_col : str
                Column in clone_df that specifies counts.
                Default none assumes count of 1 cell for each row.
            subset_ind : None or np.ndarray with partial index of df, optional
                Provides option to tally counts only within a subset of df, but to maintain the clustering
                of all individuals. Allows for one clustering of pooled TCRs,
                but tallying/testing within a subset (e.g. participants or conditions)
            hclust_method : str
                Method for hierarchical clustering, passed to the scipy.clustering.hierarchy
                linkage function.
            optimal_ordering : bool
                Flag passed to the scipy.clustering.hierarchy linkage function to improve
                visual tree layout. Can be slow for large trees.
            test_method : str or None
                Specifies Fisher's exact test ("fishers"), Chi-squared ("chi2") or
                Cochran-Mantel-Haenszel test ("chm") for testing.
    default_member_summ_kwargs: dict
        kwargs dictionary for (tcrdist.rep_diff.member_summ)

            Return additional summary info about each result (row)) based on the
            members of the cluster. This is helpful for preparing strings 
            to add to the tooltip in hierdiff.plot_hclust_props.

            res_df : pd.DataFrame [nclusters x result cols]
                Returned from neighborhood_diff or hcluster_diff
            clone_df : pd.DataFrame [nclones x metadata]
                Contains metadata for each clone.
            key_col : str
                Column in res_df that specifies the iloc of members in the clone_df
            count_col : str
                Column in clone_df that specifies counts.
                Default none assumes count of 1 cell for each row.
            addl_cols : list
                Columns to summarize
            addl_n : int
                Number of top N clones to include in the summary of
                each cluster.
    default_plot_hclust_props : dict
        kwargs dictionary for (hierdiff.plot_hclust_props)

            Plot tree of linkage-based hierarchical clustering, with nodes colored using stacked bars
            representing proportion of cluster members associated with specific conditions. Nodes also optionally
            annotated with pvalue, number of members or cluster ID.
            
            Z : linkage matrix
                Result of calling sch.linkage on a compressed pair-wise distance matrix
            res : pd.DataFrame
                Result from calling hcluster_diff, with observed/frequencies and p-values for each node
            alpha_col : str
                Column in res to use for 'alpha' annotation
            alpha : float
                Threshold for plotting the stacked bars and annotation
            colors : tuple of valid colors
                Used for stacked bars of conditions at each node
            prune_col : str/column in res
                Column of res that indicates whether a result/node can be pruned from the tree.
                The tree will not print any pruned nodes that only contain other pruned nodes.
    """
    import os
    import pandas as pd
    import numpy as np
    import warnings
    from tcrsampler.sampler import TCRsampler
    from palmotif import compute_pal_motif, svg_logo
    from tcrdist.adpt_funcs import get_centroid_seq, get_centroid_seq_alpha
    from tcrdist.summarize import _select
    from tcrdist.repertoire import TCRrep
    from tcrdist.rep_diff import hcluster_diff, member_summ
    from tcrdist.summarize import _select
    from tcrdist.pgen import OlgaModel
    from palmotif import compute_pal_motif, svg_logo
    from hierdiff import plot_hclust_props
    from numpy.random import randint
    from tcrdist.diversity import generalized_simpsons_entropy
    from tcrdist.diversity import fuzzy_diversity
    

    # Load clone_df directly from input object
    if default_hcluster_diff_kwargs['clone_df'] is None:
        default_hcluster_diff_kwargs['clone_df'] = getattr(tcrrep, 'clone_df')

    # Get appropirate pwmat_str (pw_beta, pw_alpha, or one of the CDRs, e.g., pw_cdr3_b_aa)
    if 'alpha' in tcrrep.chains:
        pwmat_str = pwmat_str_a
    if 'beta' in tcrrep.chains:
        pwmat_str = pwmat_str_b


    # Get appropirate pairwise matrix
    
    if default_hcluster_diff_kwargs['pwmat'] is None:
        if verbose : print(f"pwmat set with {pwmat_str}")
        default_hcluster_diff_kwargs['pwmat'] = getattr(tcrrep, pwmat_str)
    else: 
        if verbose : print("pwmat was directly provided as a kwarg")
    """Handle the Fact that 2 or more catagoorical levels are need to run hierdiff"""
    
    x_cols = default_hcluster_diff_kwargs['x_cols']
    if x_cols is None:
        tcrrep.clone_df['dummy'] = \
            [['X1','X2'][randint(2)] for x in range(tcrrep.clone_df.shape[0])]
        default_hcluster_diff_kwargs['x_cols'] = ['dummy']
        default_hcluster_diff_kwargs['test_method'] = 'fishers' 
        warnings.warn(f"Because x_cols was None, setting random dummy values, and using {default_hcluster_diff_kwargs['test_method']}\n", stacklevel=2)
    
    elif tcrrep.clone_df[x_cols].nunique()[0] == 2: 
        default_hcluster_diff_kwargs['test_method'] = 'fishers' 
    
    elif tcrrep.clone_df[x_cols].nunique()[0] > 2: 
        default_hcluster_diff_kwargs['test_method'] = 'chi2'
    
    elif tcrrep.clone_df[x_cols].nunique()[0] < 2 : 
        tcrrep.clone_df['dummy'] = \
            [['X1','X2'][randint(2)] for x in range(tcrrep.clone_df.shape[0])]
        default_hcluster_diff_kwargs['x_cols'] = ['dummy']
        default_hcluster_diff_kwargs['test_method'] = 'fishers' 
        warnings.warn(f"Because x_cols was None, setting random dummy values, and using {default_hcluster_diff_kwargs['test_method']}\n", stacklevel=2)

    """ Run hcluster_df """
    
    bar = IncrementalBar(f'Run hcluster_diff :', max = 2, suffix='%(percent)d%%')
    bar.next()
    tcrrep.hcluster_df, tcrrep.Z = hcluster_diff(**default_hcluster_diff_kwargs)
    bar.next(); bar.finish()

    tcrrep.hcluster_df['prune'] = tcrrep.hcluster_df['K_neighbors'].apply(lambda x: 1 if (x < prune) else 0)#
    
    """ Do Basic Summary """
    mean_distance_      = list()
    percentage_node_25_ = list()
    percentage_node_50_ = list()
    percentage_node_75_ = list()
    n_rows = tcrrep.hcluster_df.shape[0]
    bar = IncrementalBar(f'Evaluate Clusters :', max = n_rows, suffix='%(percent)d%%')
    for i,r in tcrrep.hcluster_df.iterrows():
        bar.next()
        # <dfnode> is dataframe with all the clones at a given tree node
        dfnode = tcrrep.clone_df.iloc[r['neighbors_i'],]
        # <pwnod> is dataframe with all the clones at a given tree node
        pwnode = getattr(tcrrep, pwmat_str)[r['neighbors_i'],:][:,r['neighbors_i']]
        # get the non-diaganol entries.
        node_non_diag_entries = pwnode[~np.eye(pwnode.shape[0],dtype=bool)]
        # Compute the mean distance at the node 
        mean_distance_.append(str(round(node_non_diag_entries.mean(),1)))

        percentage_node_25 = 100*(node_non_diag_entries < 25).sum() / (node_non_diag_entries.size)
        percentage_node_50 = 100*(node_non_diag_entries < 50).sum() / (node_non_diag_entries.size)#100*((pwnode < 50).sum() - pwnode.shape[0]) / (pwnode.shape[0] * pwnode.shape[1])
        percentage_node_75 = 100*(node_non_diag_entries < 75).sum() / (node_non_diag_entries.size)#100*((pwnode < 100).sum() - pwnode.shape[0]) / (pwnode.shape[0] * pwnode.shape[1])
        percentage_node_25_.append(f"{round(percentage_node_25,1)}%")
        percentage_node_50_.append(f"{round(percentage_node_50,1)}%")
        percentage_node_75_.append(f"{round(percentage_node_75,1)}%")
    bar.next(); bar.finish()
    tcrrep.hcluster_df['mean_dist']   = mean_distance_
    tcrrep.hcluster_df['pct_dist_25'] =	percentage_node_25_
    tcrrep.hcluster_df['pct_dist_50'] =	percentage_node_50_
    tcrrep.hcluster_df['pct_dist_75'] =	percentage_node_75_
    
    """
    By default, treat each clone as a single entity if 'count_col' is single 
    """
    if default_member_summ_kwargs['count_col'] == 'single':
        single = True
        print("MAKING 'single' variable")
        tcrrep.hcluster_df['single'] = 1

    if single:
        tcrrep.clone_df['single'] = 1
        default_member_summ_kwargs['count_col'] = 'single'


    """ member_summ"""
    tcrrep.res_summary = \
        member_summ(res_df = tcrrep.hcluster_df,clone_df = tcrrep.clone_df, **default_member_summ_kwargs)


    tcrrep.hcluster_df_detailed = pd.concat([tcrrep.hcluster_df.copy(), tcrrep.res_summary.copy()], axis = 1)
    
    """ Add diversity stats"""
    tcrrep.clone_df['single'] = 1
    if single:
        fdiv75 = lambda ind : fuzzy_diversity(tcrrep.clone_df.iloc[ind,:]['single'], getattr(tcrrep, pwmat_str)[ind,:][:,ind],order=2, threshold=75)
        fdiv50  = lambda ind : fuzzy_diversity(tcrrep.clone_df.iloc[ind,:]['single'], getattr(tcrrep, pwmat_str)[ind,:][:,ind],order=2, threshold=50)
        fdiv25  = lambda ind : fuzzy_diversity(tcrrep.clone_df.iloc[ind,:]['single'], getattr(tcrrep, pwmat_str)[ind,:][:,ind],order=2, threshold=25)
    else:
        fdiv75 = lambda ind : fuzzy_diversity(tcrrep.clone_df.iloc[ind,:]['count'], getattr(tcrrep, pwmat_str)[ind,:][:,ind],order=2, threshold=75)
        fdiv50  = lambda ind : fuzzy_diversity(tcrrep.clone_df.iloc[ind,:]['count'], getattr(tcrrep, pwmat_str)[ind,:][:,ind],order=2, threshold=50)
        fdiv25  = lambda ind : fuzzy_diversity(tcrrep.clone_df.iloc[ind,:]['count'], getattr(tcrrep, pwmat_str)[ind,:][:,ind],order=2, threshold=25)
    tcrrep.hcluster_df_detailed['fuzzy_simpson_diversity_25']  = [str(round(fdiv25(ind),2)) for ind in tcrrep.hcluster_df_detailed.neighbors_i.to_list()]
    tcrrep.hcluster_df_detailed['fuzzy_simpson_diversity_50']  = [str(round(fdiv50(ind),2)) for ind in tcrrep.hcluster_df_detailed.neighbors_i.to_list()]
    tcrrep.hcluster_df_detailed['fuzzy_simpson_diversity_75'] = [str(round(fdiv75(ind),2)) for ind in tcrrep.hcluster_df_detailed.neighbors_i.to_list()]

    """Optional Add SVGs to hcluster_detailed"""
    if 'beta' in tcrrep.chains:
        _tcrsampler_svgs(tcrrep = tcrrep,
            default_background = None, 
            default_background_if_missing = None,
            cdr3_name = 'cdr3_b_aa',
            pwmat_str = pwmat_str_b, 
            chain = 'beta', 
            gene_names = ['v_b_gene','j_b_gene'],
            combine_olga = combine_olga)
    
    if 'alpha' in tcrrep.chains:
        _tcrsampler_svgs(tcrrep = tcrrep,
            default_background = None, 
            default_background_if_missing = None,
            cdr3_name = 'cdr3_a_aa',
            pwmat_str = pwmat_str_a, 
            chain = 'alpha', 
            gene_names = ['v_a_gene','j_a_gene'],
            combine_olga = combine_olga)


    """ Plot """
    html = plot_hclust_props(tcrrep.Z,
            res=tcrrep.hcluster_df_detailed,
            prune_col = 'prune',
            **default_plot_hclust_props)

    """ Write File """		
    with open(html_name, 'w') as fh:
        print(f"WRITING {html_name}")
        fh.write(html)


def report_kwargs(keyword_arguments):
    return ", ".join([f"{k} : {v}" for k,v in keyword_arguments.items()])

def _tcrsampler_svgs(
    tcrrep,
    default_background = None, 
    default_background_if_missing = None,
    cdr3_name = 'cdr3_b_aa',
    pwmat_str = 'pw_cdr3_b_aa', 
    chain = 'beta', 
    gene_names = ['v_b_gene','j_b_gene'],
    combine_olga = False,
    verbose = True):
    """
    Breath. What does this do?
    
    Given a TCRrep instance, this function samples a background repertoire
    using TCRsampler and makes svg-logos using palmotif. 

    This function doesn't return anything it. It needs to access 
    attribute values of a TCRrep (tcrrep) instance and 
    it modifies th etcrrep in place adding svgs and stats colums 
    to .hcluster_df_detailed DataFrame. TODO: could just output a dataframe 
    which would then just be concattenated.

    ONLY WORKS WITH _BETA using defaults:

    Notes
    -----
    Note: TCRSampler.build_background() accepts kwargs, we've set these as fixed as most user 
    won't know what these do and won't need to change them.
            max_rows : int
                Maximum clones per v,j pair (per subject)
            stratify_by_subject : bool
                If True, max_rows will apply to v,j,subject. If False, max_rows applies to v,j
            use_frequency : bool
                If True, uses frequency for ranking rows. If False, uses raw counts.
            make_singleton : bool
                If True, background is still sorted by frequency or counts, 
                but final fequency and counts values are overridden
                and set to 1.
    """  
    from tcrsampler.sampler import TCRsampler
    from palmotif import compute_pal_motif, svg_logo
    import pandas as pd
    from tcrdist.summarize import _select
    
    if chain == 'alpha' and tcrrep.organism == "mouse":
        # Here we enforce the rule that alpha-mouse cannot use an olga-sampler
        # TODO: This should be removed as soon as TCRsampler can be updated with a valid 
        # mouse-alpha simulated background.
        combine_olga = False
    
    # _default_sampler returns a TCRSampler based on organism and chain
    if verbose: print(f"INITIALIZING A TCRSAMPLER")
    print(tcrrep.organism, chain)
    t = _default_sampler(organism = tcrrep.organism, chain = chain)(
        default_background = default_background, 
        default_background_if_missing = default_background_if_missing)

    build_kargs = { 'max_rows' : 100,
                    'stratify_by_subject' : True,
                    'use_frequency' : True,
                    'make_singleton' : False}

    build_kargs_olga = {   'max_rows' : 1000,
                        'stratify_by_subject' : False,
                        'use_frequency' : True,
                        'make_singleton' : False}
    
    if verbose: print(f"BUILDING A DEEPER BACKGROUND {report_kwargs(build_kargs)}")

    t.build_background(**build_kargs)
    # Olga Sampler
    
    if combine_olga:
        t_olga = _default_sampler_olga(chain = chain, organism = tcrrep.organism)()
        if verbose: print(f"BUILDING A DEEPER BACKGROUND {report_kwargs(build_kargs_olga)}")
        t.build_background(**build_kargs_olga )

        olga_model = {
            ('beta','human')  : OlgaModel(recomb_type="VDJ", chain_folder = "human_T_beta"),
            ('alpha','human') : OlgaModel(recomb_type="VJ",  chain_folder = "human_T_alpha"),
            ('beta','mouse')  : OlgaModel(recomb_type="VDJ", chain_folder = "mouse_T_beta")
            }[(chain, tcrrep.organism)]



    if 'prune' not in tcrrep.hcluster_df.columns:
        if verbose: print("NO PRUNE COLUMNS USED ALL SET TO 0")
        tcrrep.hcluster_df['prune'] = 0
    
    print("ITERATE THROUGH CLUSTERS")
    svgs = list()
    svgs_raw = list()
    reference_unique = list()
    reference_unique_olga= list()
    reference_size = list()
    reference_size_olga = list()
    percent_missing_sampler = list()
    percent_missing_sampler_olga = list()
    n_rows = tcrrep.hcluster_df.shape[0]
    
    
    bar = IncrementalBar(f'Make {chain} SVGs :', max = n_rows, suffix='%(percent)d%%')
    for i,r in tcrrep.hcluster_df.iterrows():
        bar.next()
        if r['prune'] == 0:
            # <dfnode> is dataframe with all the clones at a given tree node
            dfnode   = tcrrep.clone_df.iloc[r['neighbors_i'],].copy()
            # <pwnode> Pairwise Matrix for node sequences
            pwnode   = getattr(tcrrep, pwmat_str)[r['neighbors_i'],:][:,r['neighbors_i']].copy()
            iloc_idx = pwnode.sum(axis = 0).argmin()
            centroid = dfnode[cdr3_name].to_list()[iloc_idx]
    
            # Compute gene usage at the node  
            # Convert to allele_01
            for gene_name in gene_names:
                dfnode[gene_name] = dfnode[gene_name].apply(lambda x : allele_01(x)) 
                      
            gene_usage = dfnode.groupby(gene_names).size() # e.g., ['v_b_gene','j_b_gene']
            gene_usage_tuples = gene_usage.reset_index().to_dict('split')['data']
            # Given gene usage use the <t> a TCRsampler instance to get background seqs
            
            # Adjust depth for small nodes
            adjust_depth = 10*round(10 / dfnode.shape[0])
            if adjust_depth < 10:
                adjust_depth = 10

            sampled_rep = t.sample( gene_usage_tuples,
                            flatten = True, depth = adjust_depth * 10)
            
            # Only keep the non-none sequences
            sampled_rep  = [x for x in sampled_rep if x is not None]
            # < missing_gene > Count the percentage missing, sampler returns none when no v,j pair is present
            expected_depth = dfnode.shape[0] * adjust_depth * 10
            recovered_depth = len(sampled_rep)
            percent_missing = round(100* (1- (recovered_depth / expected_depth)), 1)
                   
            percent_missing_sampler.append(f"{percent_missing}%")           
            reference_unique.append(str(pd.Series(sampled_rep).nunique()))
            reference_size.append(str(pd.Series(sampled_rep).count()))

            if combine_olga:
                # We modified Olga source code slightly, such that we simulated sequences 
                # with a given V,J gene usage
                # OLD METHOD WHERE WE ACTUALLY SAMPLED, slower but can go much deeper. I don't think one rare sequence however, really make a big difference.
                #flatten = lambda l: [item for sublist in l for item in sublist]
                #sampled_rep_olga = [olga_model.gen_cdr3s(allele_01(v),allele_01(j),n*adjust_depth*10) for v,j,n in gene_usage_tuples]
                #sampled_rep_olga = [x for x in flatten(sampled_rep_olga) if x is not None]
                sampled_rep_olga = t_olga.sample( gene_usage_tuples,
                            flatten = True, depth = adjust_depth * 10)

                sampled_rep_olga = [x for x in sampled_rep_olga if x is not None]
                
                expected_depth = dfnode.shape[0] * adjust_depth * 10
                recovered_depth = len(sampled_rep_olga)
                percent_missing_olga = round(100* (1- (recovered_depth / expected_depth)), 1)
                
                percent_missing_sampler_olga.append(f"{percent_missing_olga}%")            
                reference_unique_olga.append(str(pd.Series(sampled_rep_olga).nunique()))
                reference_size_olga.append(str(pd.Series(sampled_rep_olga).count()))
                
                # HERE WE COMBINE INTO A SINGLE BACKGROUND:
                sampled_rep = sampled_rep + sampled_rep_olga


            # Get motif matrix and motif stats
            motif, stat = compute_pal_motif(
                            seqs = _select(df = tcrrep.clone_df, 
                                        iloc_rows = r['neighbors_i'], 
                                        col = cdr3_name),
                            refs = sampled_rep, 
                            centroid = centroid)

            svgs.append(svg_logo(motif, return_str= True))

            # repeaat without references
            raw_motif, raw_stat = compute_pal_motif(
                seqs = _select(df = tcrrep.clone_df, 
                            iloc_rows = r['neighbors_i'], 
                            col = cdr3_name),
                centroid = centroid)
            # Convert the motif matrix into an svg_logo, append to list
            svgs_raw.append(svg_logo(raw_motif, return_str= True))
        else: 
            # If prune column is 1 don't go to the trouble of sampling and generating seqs
            svgs.append("PRUNE")
            svgs_raw.append("PRUNE")
            reference_size.append("PRUNE")
            reference_unique.append("PRUNE")
            percent_missing_sampler.append("PRUNE")
            percent_missing_sampler_olga.append("PRUNE")            
            reference_unique_olga.append("PRUNE") 
            reference_size_olga.append("PRUNE") 

    bar.next(); bar.finish()

    # The standard svg_ includes background, whereas raw has no background 
    tcrrep.hcluster_df_detailed[f'svg_{chain}'] 		= svgs
    tcrrep.hcluster_df_detailed[f'svg_raw_{chain}'] 	= svgs_raw
    tcrrep.hcluster_df_detailed[f'ref_size_{chain}'] 	= reference_size
    tcrrep.hcluster_df_detailed[f'ref_unique_{chain}'] 	= reference_unique
    tcrrep.hcluster_df_detailed[f'percent_missing_{chain}']  = percent_missing_sampler   
    if combine_olga:
        tcrrep.hcluster_df_detailed[f'ref_size_olga_{chain}'] 	      = reference_size_olga
        tcrrep.hcluster_df_detailed[f'ref_unique_olga_{chain}'] 	  = reference_unique_olga
        tcrrep.hcluster_df_detailed[f'percent_missing_olga_{chain}']  = percent_missing_sampler_olga  
    
    return True



def _default_sampler(organism = 'human', chain = 'beta'):
    assert organism in ['human', 'mouse']
    assert chain in ['beta','alpha']

    default_tcrsampler_generator = {
        ('human','beta'): 
            _default_tcrsampler_human_beta,
        ('human','alpha'): 
            _default_tcrsampler_human_alpha,
        ('mouse','beta'): 
            _default_tcrsampler_mouse_beta,
        ('mouse','alpha'): 
            _default_tcrsampler_mouse_alpha, 
        }[(organism, chain)]
    
    return default_tcrsampler_generator

def _default_sampler_olga(organism = 'human', chain = 'beta'):
    assert organism in ['human', 'mouse']
    assert chain in ['beta','alpha']
    if organism == "mouse" and chain == "alpha":
        raise KeyError("No currenlty available mouse-alpha chain background form olga")

    default_tcrsampler_generator = {
        ('human','beta'): 
            _default_tcrsampler_olga_human_beta,
        ('human','alpha'): 
            _default_tcrsampler_olga_human_alpha,
        ('mouse','beta'): 
            _default_tcrsampler_olga_mouse_beta
        }[(organism, chain)]
    
    return default_tcrsampler_generator


def _default_tcrsampler_olga_human_beta(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background =  'olga_human_beta_t.sampler.tsv'
        
    if default_background_if_missing is None:
        default_background_if_missing ='olga_sampler.zip'
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_olga_human_alpha(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background =  'olga_human_alpha_t.sampler.tsv'
        
    if default_background_if_missing is None:
        default_background_if_missing ='olga_sampler.zip'
    
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_olga_mouse_beta(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background =  'olga_mouse_beta_t.sampler.tsv'
        
    if default_background_if_missing is None:
        default_background_if_missing ='olga_sampler.zip'
    
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t


def _default_tcrsampler_human_beta(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default human beta sampler 'britanova_human_beta_t_cb.tsv.sampler.tsv'

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background = 'britanova_human_beta_t_cb.tsv.sampler.tsv'
        
    if default_background_if_missing is None:
        default_background_if_missing ='britanova_human_beta_t_cb.tsv.sampler.tsv.zip'
    
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_human_alpha(default_background = None, default_background_if_missing=None ):
    """
    Responsible for providing the default human alpha sampler 'ruggiero_human_alpha_t.tsv.sampler.tsv'
    """
    from tcrsampler.sampler import TCRsampler
    if default_background is None:
        default_background = 'ruggiero_human_alpha_t.tsv.sampler.tsv'
    if default_background_if_missing is None:
        default_background_if_missing =  'ruggiero_human_alpha_t.tsv.sampler.tsv.zip'
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_mouse_beta(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default mouse beta sampler

    Returns
    -------
    t : tcrsampler.sampler.TCRsampler 
    """
    from tcrsampler.sampler import TCRsampler

    if default_background is None:
        default_background = 'ruggiero_mouse_beta_t.tsv.sampler.tsv'
    if default_background_if_missing is None:
        default_background_if_missing =  'ruggiero_mouse_sampler.zip'

    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def _default_tcrsampler_mouse_alpha(default_background = None, default_background_if_missing=None):
    """
    Responsible for providing the default mouse alpha sampler
    """
    from tcrsampler.sampler import TCRsampler
    
    if default_background is None:
        default_background = 'ruggiero_mouse_alpha_t.tsv.sampler.tsv'
    if default_background_if_missing is None:
        default_background_if_missing =  'ruggiero_mouse_sampler.zip'
    
    print(default_background)

    try: 
        t = TCRsampler(default_background=default_background)
    except OSError:
        t = TCRsampler()
        t.download_background_file(default_background_if_missing)
        t = TCRsampler(default_background=default_background)
    return t

def allele_01(genename):
    """
    >>> allele_01('TRBV19*02')
    'TRBV19*01'
    """
    g,a = re.match(pattern = '(.*)([0-9])$', string= genename).groups()
    return(f"{g}1")


import pytest



def test_import_vdjtools_beta_w_validation():
    import pandas as pd
    import numpy as np
    import os
    from tcrdist.paths import path_to_base
    from tcrdist.vdjtools_funcs import import_vdjtools
    from tcrdist.repertoire import TCRrep

    # Reformat vdj_tools input format for tcrdist3
    vdj_tools_file_beta = os.path.join(path_to_base, 'tcrdist','data','formats','vdj.M_15_CD8_beta.clonotypes.TRB.txt.gz')
    df_beta = import_vdjtools(   vdj_tools_file = vdj_tools_file_beta ,
                        chain = 'beta',
                        organism = 'human',
                        db_file = 'alphabeta_gammadelta_db.tsv',
                        validate = True)
    assert np.all(df_beta.columns == ['count', 'freq', 'cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'cdr3_b_nucseq','valid_v', 'valid_j', 'valid_cdr3'])
    
    # Can be directly imported into a TCRrep instance.
    tr = TCRrep(
        cell_df = df_beta[['count', 'freq', 'cdr3_b_aa', 'v_b_gene', 'j_b_gene']], 
        chains = ['beta'], 
        organism = 'human', 
        compute_distances = False)


def test_import_vdjtools_beta_no_validation():
    import pandas as pd
    import numpy as np
    import os
    from tcrdist.paths import path_to_base
    from tcrdist.vdjtools_funcs import import_vdjtools
    vdj_tools_file_beta = os.path.join(path_to_base, 'tcrdist','data','formats','vdj.M_15_CD8_beta.clonotypes.TRB.txt.gz')
    df_beta = import_vdjtools(   vdj_tools_file = vdj_tools_file_beta ,
                        chain = 'beta',
                        organism = 'human',
                        db_file = 'alphabeta_gammadelta_db.tsv',
                        validate = False)
    assert np.all(df_beta.columns == ['count', 'freq', 'cdr3_b_aa', 'v_b_gene', 'j_b_gene', 'cdr3_b_nucseq','valid_v', 'valid_j', 'valid_cdr3'])
    #assert False in df_beta.valid_cdr3.to_list()
    #assert False in df_beta.valid_v.to_list()
    #assert False in df_beta.valid_j.to_list()


def test_import_vdjtools_alpha_w_validation():
    import pandas as pd
    import numpy as np
    import os
    from tcrdist.paths import path_to_base
    from tcrdist.vdjtools_funcs import import_vdjtools
    vdj_tools_file_alpha = os.path.join(path_to_base, 'tcrdist','data','formats','vdj.M_15_CD8_alpha.clonotypes.TRA.txt.gz')
    df_alpha = import_vdjtools( vdj_tools_file = vdj_tools_file_alpha,
                        chain = 'alpha',
                        organism = 'human',
                        db_file = 'alphabeta_gammadelta_db.tsv',
                        validate = True)
    assert np.all(df_alpha.columns == ['count', 'freq', 'cdr3_a_aa', 'v_a_gene', 'j_a_gene', 'cdr3_a_nucseq','valid_v', 'valid_j', 'valid_cdr3'])

def test_import_vdjtools_alpha_no_validation():
    import pandas as pd
    import numpy as np
    import os
    from tcrdist.paths import path_to_base
    from tcrdist.vdjtools_funcs import import_vdjtools
    vdj_tools_file_alpha = os.path.join(path_to_base, 'tcrdist','data','formats','vdj.M_15_CD8_alpha.clonotypes.TRA.txt.gz')
    df_alpha = import_vdjtools( vdj_tools_file = vdj_tools_file_alpha,
                        chain = 'alpha',
                        organism = 'human',
                        db_file = 'alphabeta_gammadelta_db.tsv',
                        validate = False)
    assert np.all(df_alpha.columns == ['count', 'freq', 'cdr3_a_aa', 'v_a_gene', 'j_a_gene', 'cdr3_a_nucseq','valid_v', 'valid_j', 'valid_cdr3'])
    #assert False in df_alpha.valid_cdr3.to_list()
    #assert False in df_alpha.valid_v.to_list()
    #assert False in df_alpha.valid_j.to_list()

